/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2006 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "IconTexture.h"
#include "ExtractResource.h"

#include <shlobj.h>

extern UINT (WINAPI *_PrivateExtractIcons)(LPCTSTR,int,int,int,HICON*,UINT*,UINT,UINT);

IconTexture::IconTexture() :
	memoryDC(0),
	memoryBitmap(0),

	memoryHoverDC(0),
	memoryHoverBitmap(0),

	memoryPressedDC(0),
	memoryPressedBitmap(0),

	transparencyNeeded(false),

	useAlphaMap(false),

	mustUseAlpha(false),
	hovermustUseAlpha(false),
	pressedmustUseAlpha(false),

	useHover(false),
	usePressed(false)
{
}

IconTexture::~IconTexture()
{
	if(memoryBitmap)
	{
		memoryBitmap = (HBITMAP) SelectObject(memoryDC, memoryBitmap);
		DeleteObject(memoryBitmap);
	}
	if(memoryDC)
		DeleteDC(memoryDC);

	if(memoryHoverBitmap)
	{
		memoryHoverBitmap = (HBITMAP) SelectObject(memoryHoverDC, memoryHoverBitmap);
		DeleteObject(memoryHoverBitmap);
	}
	if(memoryHoverDC)
		DeleteDC(memoryHoverDC);

	if(memoryPressedBitmap)
	{
		memoryPressedBitmap = (HBITMAP) SelectObject(memoryPressedDC, memoryPressedBitmap);
		DeleteObject(memoryPressedBitmap);
	}
	if(memoryPressedDC)
		DeleteDC(memoryPressedDC);
}

void IconTexture::configure(const string &uniqueKey, const string &groupKey, bool alphaMap)
{
	NameValuePair borderValues[] = 
	{
		{ ".none", 0 },
		{ "none", 0 },
		{ "raised", 1 },
		{ "sunken", 2 },
		{ "etched", 3 },
		{ "bump", 4 },
		{ 0, 0 }
	};

	CIconExtractor m_IconExtractor;

	COLORREF hueColor, hoverhueColor, pressedhueColor;
	int hueIntensity, hoverhueIntensity, pressedhueIntensity;
	int luminanceIntensity, hoverluminanceIntensity, pressedluminanceIntensity;
	int saturationIntensity, hoversaturationIntensity, pressedsaturationIntensity;

	COLORREF mixColor, hovermixColor, pressedmixColor;
	int mixIntensity, hovermixIntensity, pressedmixIntensity;

	//Always Needed Default Values
	//Normal (for Hover) and Hover (for Pressed)
	//--------------------------------------------
	hueColor = hoverhueColor = Getx2RCColor(uniqueKey, groupKey, "HueColor", RGB(255, 255, 255));
	hueIntensity = hoverhueIntensity = Getx2RCInt(uniqueKey, groupKey, "HueIntensity", 0, 0, 100);
	luminanceIntensity = hoverluminanceIntensity = Getx2RCInt(uniqueKey, groupKey, "LuminanceIntensity", 0, -100, 100);
	saturationIntensity = hoversaturationIntensity = Getx2RCInt(uniqueKey, groupKey, "SaturationIntensity", 100, 0, 200);
	
	mixColor = hovermixColor = Getx2RCColor(uniqueKey, groupKey, "MixColor", RGB(255, 255, 255));
	mixIntensity = hovermixIntensity = Getx2RCInt(uniqueKey, groupKey, "MixIntensity", 0, 0, 100);
	
	borderMethod = hoverborderMethod = Getx2RCNamedValue(uniqueKey, groupKey, "BorderMethod", borderValues, 0);
	//--------------------------------------------

	useAlphaMap = alphaMap;

	int ExtractIconSize = Getx2RCInt(uniqueKey, groupKey, "ExtractIconSize", 32, 16, 256);
	bool OldStyle = Getx2RCBool(uniqueKey, groupKey, "OldStyleIconExtraction");

	string iconImageFile = Getx2RCString(uniqueKey, groupKey, "Image", "");

	if ( !iconImageFile.empty() )
	{
		CString m_strPath = iconImageFile.c_str();
		m_strPath.MakeLower();

		HICON noalphaMapIcon = NULL;
		int maxIconSize = ExtractIconSize;

		if ( m_strPath.Find(".ico") != -1 )
		{
			noalphaMapIcon = m_IconExtractor.ExtractIconFromIcoEx(m_strPath, memoryBitmap, maxIconSize);
		}
		else
		{
			int location = 0;
			if ( !OldStyle && m_strPath.Find(",") != -1 )
			{
				location = atoi(m_strPath.Right(m_strPath.GetLength()-(m_strPath.ReverseFind(',')+1)));
				m_strPath = m_strPath.Left(m_strPath.ReverseFind(','));
			}

			if ( !OldStyle && (m_strPath.Find(".exe") != -1 || m_strPath.Find(".dll") != -1) )
			{
				noalphaMapIcon = m_IconExtractor.ExtractIconFromExeEx(m_strPath, location, memoryBitmap, maxIconSize);
			}
			else if ( !OldStyle && m_strPath.Find(".lnk") != -1 )
			{
				 GetIconFromLnkTarget(m_strPath, ExtractIconSize, 0, noalphaMapIcon);	
			}
			else
			{
				SHFILEINFO shFileInfo;
				if ( OldStyle )
					SHGetFileInfo(iconImageFile.c_str(), NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_ICON | (ExtractIconSize < 32 ? SHGFI_SMALLICON : 0) );
				else
					SHGetFileInfo(m_strPath, NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_ICON | (ExtractIconSize < 32 ? SHGFI_SMALLICON : 0) );
					
				noalphaMapIcon = shFileInfo.hIcon;
			}
		}

		iconSize = maxIconSize;

		if ( memoryBitmap != NULL || noalphaMapIcon != NULL )
			CreateDCFromSource(memoryDC, memoryBitmap, hueColor, hueIntensity, luminanceIntensity, saturationIntensity, mixColor, mixIntensity, iconSize, noalphaMapIcon);

		if ( noalphaMapIcon == NULL || IsOS( OS_XP ) )
			mustUseAlpha = true;
		if ( noalphaMapIcon != NULL )
			DestroyIcon( noalphaMapIcon );
	}

	iconImageFile = Getx2RCString(uniqueKey, groupKey, "HoverImage", "");

	if ( !iconImageFile.empty() )
	{
		useHover = true;

		hoverhueColor = Getx2RCColor(uniqueKey, groupKey, "HoverHueColor", hueColor);
		hoverhueIntensity = Getx2RCInt(uniqueKey, groupKey, "HoverHueIntensity", hueIntensity, 0, 100);
		hoverluminanceIntensity = Getx2RCInt(uniqueKey, groupKey, "HoverLuminanceIntensity", luminanceIntensity, -100, 100);
		hoversaturationIntensity = Getx2RCInt(uniqueKey, groupKey, "HoverSaturationIntensity", saturationIntensity, 0, 200);

		hovermixColor = Getx2RCColor(uniqueKey, groupKey, "HoverMixColor", mixColor);
		hovermixIntensity = Getx2RCInt(uniqueKey, groupKey, "HoverMixIntensity", mixIntensity, 0, 100);

		hoverborderMethod = Getx2RCNamedValue(uniqueKey, groupKey, "HoverBorderMethod", borderValues, borderMethod);

		CString m_strPath = iconImageFile.c_str();
		m_strPath.MakeLower();

		HICON noalphaMapIcon = NULL;
		int maxIconSize = ExtractIconSize;

		if ( m_strPath.Find(".ico") != -1 )
		{
			noalphaMapIcon = m_IconExtractor.ExtractIconFromIcoEx(m_strPath, memoryHoverBitmap, maxIconSize);
		}
		else
		{
			int location = 0;
			if ( !OldStyle && m_strPath.Find(",") != -1 )
			{
				location = atoi(m_strPath.Right(m_strPath.GetLength()-(m_strPath.ReverseFind(',')+1)));
				m_strPath = m_strPath.Left(m_strPath.ReverseFind(','));
			}

			if ( !OldStyle && (m_strPath.Find(".exe") != -1 || m_strPath.Find(".dll") != -1) )
			{
				noalphaMapIcon = m_IconExtractor.ExtractIconFromExeEx(m_strPath, location, memoryHoverBitmap, maxIconSize);
			}
			else if ( !OldStyle && m_strPath.Find(".lnk") != -1 )
			{
				GetIconFromLnkTarget(m_strPath, ExtractIconSize, 1, noalphaMapIcon);	
			}
			else
			{
				SHFILEINFO shFileInfo;
				if ( OldStyle )
					SHGetFileInfo(iconImageFile.c_str(), NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_ICON | (ExtractIconSize < 32 ? SHGFI_SMALLICON : 0) );
				else
					SHGetFileInfo(m_strPath, NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_ICON | (ExtractIconSize < 32 ? SHGFI_SMALLICON : 0) );
				
				noalphaMapIcon = shFileInfo.hIcon;
			}
		}
		
		hovericonSize = maxIconSize;
			
		if ( memoryHoverBitmap != NULL || noalphaMapIcon != NULL )
			CreateDCFromSource(memoryHoverDC, memoryHoverBitmap, hoverhueColor, hoverhueIntensity, hoverluminanceIntensity, hoversaturationIntensity, hovermixColor, hovermixIntensity, hovericonSize, noalphaMapIcon);

		if ( noalphaMapIcon == NULL || IsOS( OS_XP ) )
			hovermustUseAlpha = true;
		if ( noalphaMapIcon != NULL )
			DestroyIcon( noalphaMapIcon );
	}

	iconImageFile = Getx2RCString(uniqueKey, groupKey, "PressedImage", "");

	if ( !iconImageFile.empty() )
	{
		usePressed = true;

		pressedhueColor = Getx2RCColor(uniqueKey, groupKey, "PressedHueColor", hoverhueColor);
		pressedhueIntensity = Getx2RCInt(uniqueKey, groupKey, "PressedHueIntensity", hoverhueIntensity, 0, 100);
		pressedluminanceIntensity = Getx2RCInt(uniqueKey, groupKey, "PressedLuminanceIntensity", hoverluminanceIntensity, -100, 100);
		pressedsaturationIntensity = Getx2RCInt(uniqueKey, groupKey, "PressedSaturationIntensity", hoversaturationIntensity, 0, 200);

		pressedmixColor = Getx2RCColor(uniqueKey, groupKey, "PressedMixColor", hovermixColor);
		pressedmixIntensity = Getx2RCInt(uniqueKey, groupKey, "PressedMixIntensity", hovermixIntensity, 0, 100);

		pressedborderMethod = Getx2RCNamedValue(uniqueKey, groupKey, "PressedBorderMethod", borderValues, hoverborderMethod);

		CString m_strPath = iconImageFile.c_str();
		m_strPath.MakeLower();

		HICON noalphaMapIcon = NULL;
		int maxIconSize = ExtractIconSize;

		if ( m_strPath.Find(".ico") != -1 )
		{
			noalphaMapIcon = m_IconExtractor.ExtractIconFromIcoEx(m_strPath, memoryPressedBitmap, maxIconSize);
		}
		else
		{
			int location = 0;
			if ( !OldStyle && m_strPath.Find(",") != -1 )
			{
				location = atoi(m_strPath.Right(m_strPath.GetLength()-(m_strPath.ReverseFind(',')+1)));
				m_strPath = m_strPath.Left(m_strPath.ReverseFind(','));
			}

			if ( !OldStyle && (m_strPath.Find(".exe") != -1 || m_strPath.Find(".dll") != -1) )
			{
				noalphaMapIcon = m_IconExtractor.ExtractIconFromExeEx(m_strPath, location, memoryPressedBitmap, maxIconSize);
			}
			else if ( !OldStyle && m_strPath.Find(".lnk") != -1 )
			{
				GetIconFromLnkTarget(m_strPath, ExtractIconSize, 2, noalphaMapIcon);	
			}
			else
			{
				SHFILEINFO shFileInfo;
				if ( OldStyle )
					SHGetFileInfo(iconImageFile.c_str(), NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_ICON | (ExtractIconSize < 32 ? SHGFI_SMALLICON : 0) );
				else
					SHGetFileInfo(m_strPath, NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_ICON | (ExtractIconSize < 32 ? SHGFI_SMALLICON : 0) );
			
				noalphaMapIcon = shFileInfo.hIcon;
			}
		}
		
		pressediconSize = maxIconSize;
			
		if ( memoryPressedBitmap != NULL || noalphaMapIcon != NULL )
			CreateDCFromSource(memoryPressedDC, memoryPressedBitmap, pressedhueColor, pressedhueIntensity, pressedluminanceIntensity, pressedsaturationIntensity, pressedmixColor, pressedmixIntensity, pressediconSize, noalphaMapIcon);

		if ( noalphaMapIcon == NULL || IsOS( OS_XP ) )
			pressedmustUseAlpha = true;
		if ( noalphaMapIcon != NULL )
			DestroyIcon( noalphaMapIcon );
	}
}

void IconTexture::apply(HDC hDC, int x, int y, int width, int height, int transparencyMode, int imageType)
{
	if ( width <= 0 || height <= 0 )
		return;

	if ( imageType == 2 && (pressedmustUseAlpha && !useAlphaMap) )
	{
		applyOverlay(hDC, x, y, width, height, 255, 2);
		return;
	}
	else if ( imageType == 1 && (hovermustUseAlpha && !useAlphaMap) )
	{
		applyOverlay(hDC, x, y, width, height, 255, 1);
		return;
	}
	else if (imageType == 0 && (mustUseAlpha && !useAlphaMap) )
	{
		applyOverlay(hDC, x, y, width, height, 255, 0);
		return;
	}

	if ( imageType == 2 && memoryPressedBitmap == 0 )
		imageType = 1;

	if ( imageType == 1 && memoryHoverBitmap == 0 )
		imageType = 0;

	if ( imageType == 0 && memoryBitmap == 0 )
		return;

	HDC bufferDC = CreateCompatibleDC(NULL);

	BITMAPINFO alpha;
	ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
	alpha.bmiHeader.biWidth=width;      // Set size you need
	alpha.bmiHeader.biHeight=height;    // Set size you need
	alpha.bmiHeader.biPlanes=1;
	alpha.bmiHeader.biBitCount=32;      // Can be 8, 16, 32 bpp or other number
	alpha.bmiHeader.biSizeImage=0;
	alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	alpha.bmiHeader.biClrUsed= 0;
	alpha.bmiHeader.biClrImportant= 0;
	VOID *pvBits;

	HBITMAP bufferBitmap = CreateDIBSection( bufferDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

	HBITMAP oldBMP = (HBITMAP) SelectObject(bufferDC, bufferBitmap);

	SetStretchBltMode(bufferDC, STRETCH_DELETESCANS);

	if (imageType == 0)
	{
		MultiBlt(bufferDC, 0, 0, width, height,
					memoryDC, 0, 0, iconSize, iconSize,
					0, 0, 0, 0,
					MULTIBLT_STRETCH, SRCCOPY);
	}
	else if (imageType == 1)
	{
		MultiBlt(bufferDC, 0, 0, width, height,
					memoryHoverDC, 0, 0, hovericonSize, hovericonSize,
					0, 0, 0, 0,
					MULTIBLT_STRETCH, SRCCOPY);
	}
	else if (imageType == 2)
	{
		MultiBlt(bufferDC, 0, 0, width, height,
					memoryPressedDC, 0, 0, pressediconSize, pressediconSize,
					0, 0, 0, 0,
					MULTIBLT_STRETCH, SRCCOPY);
	}

	BITMAP bm;
	GetObject(bufferBitmap, sizeof(BITMAP), &bm);

	BYTE *pPixel = (LPBYTE)bm.bmBits;

	int iSize = bm.bmHeight * bm.bmWidth;

	for (int j = 0; j < iSize; ++j)
	{
		//Magic Pink
		if ( pPixel[0] == 0 && pPixel[1] == 0 && pPixel[2] == 0 && pPixel[3] == 0 )
		{
			if ( !useAlphaMap )
			{
				pPixel[0] = 255;
				pPixel[1] = 0;
				pPixel[2] = 255;
				pPixel[3] = 0;
			}
			else
			{
				pPixel[3] = 1;
			}
		}
			
		pPixel+= 4;
	}

	if ( transparencyMode == 2 || transparencyMode == 3 )
	{
		TransparentBltLS(hDC, x, y, width, height,
							bufferDC, 0, 0,
							RGB(255, 0, 255));
	}
	else
	{
		BitBlt(hDC, x, y, width, height, bufferDC, 0, 0, SRCCOPY);
	}

	RECT completeFillRect = {x, y, x+width, y+height};

	if ( (borderMethod == 1 && imageType == 0) || (hoverborderMethod == 1 && imageType == 1) || (pressedborderMethod == 1 && imageType == 2) )
		DrawEdge(hDC, &completeFillRect, EDGE_RAISED, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 2 && imageType == 0) || (hoverborderMethod == 2 && imageType == 1) || (pressedborderMethod == 2 && imageType == 2) )
		DrawEdge(hDC, &completeFillRect, EDGE_SUNKEN, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 3 && imageType == 0) || (hoverborderMethod == 3 && imageType == 1) || (pressedborderMethod == 3 && imageType == 2) )
		DrawEdge(hDC, &completeFillRect, EDGE_ETCHED, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 4 && imageType == 0) || (hoverborderMethod == 4 && imageType == 1) || (pressedborderMethod == 4 && imageType == 2) )
		DrawEdge(hDC, &completeFillRect, EDGE_BUMP, BF_RECT | BF_SOFT);

	bufferBitmap = (HBITMAP) SelectObject(bufferDC, oldBMP);
	DeleteObject(bufferBitmap);
	DeleteDC(bufferDC);
}

void IconTexture::applyOverlay(HDC hDC, int x, int y, int width, int height, int alphaValue, int imageType)
{	
	if ( width <= 0 || height <= 0 )
		return;

	if ( imageType == 2 && memoryPressedBitmap == 0 )
		imageType = 1;

	if ( imageType == 1 && memoryHoverBitmap == 0 )
		imageType = 0;

	if ( imageType == 0 && memoryBitmap == 0 )
		return;

	HDC bufferDC = CreateCompatibleDC(NULL);

	BITMAPINFO alpha;
	ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
	alpha.bmiHeader.biWidth=width;      // Set size you need
	alpha.bmiHeader.biHeight=height;    // Set size you need
	alpha.bmiHeader.biPlanes=1;
	alpha.bmiHeader.biBitCount=32;      // Can be 8, 16, 32 bpp or other number
	alpha.bmiHeader.biSizeImage=0;
	alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	alpha.bmiHeader.biClrUsed= 0;
	alpha.bmiHeader.biClrImportant= 0;
	VOID *pvBits;

	HBITMAP bufferBitmap = CreateDIBSection( bufferDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

	HBITMAP oldBMP = (HBITMAP) SelectObject(bufferDC, bufferBitmap);

	SetStretchBltMode(bufferDC, STRETCH_DELETESCANS);

	if (imageType == 0)
	{
		MultiBlt(bufferDC, 0, 0, width, height,
					memoryDC, 0, 0, iconSize, iconSize,
					0, 0, 0, 0,
					MULTIBLT_STRETCH, SRCCOPY);
	}
	else if (imageType == 1)
	{
		MultiBlt(bufferDC, 0, 0, width, height,
					memoryHoverDC, 0, 0, hovericonSize, hovericonSize,
					0, 0, 0, 0,
					MULTIBLT_STRETCH, SRCCOPY);
	}
	else if (imageType == 2)
	{
		MultiBlt(bufferDC, 0, 0, width, height,
					memoryPressedDC, 0, 0, pressediconSize, pressediconSize,
					0, 0, 0, 0,
					MULTIBLT_STRETCH, SRCCOPY);
	}

	RECT completeFillRect = {0, 0, width, height};

	if ( (borderMethod == 1 && imageType == 0) || (hoverborderMethod == 1 && imageType == 1) || (pressedborderMethod == 1 && imageType == 2) )
		DrawEdge(bufferDC, &completeFillRect, EDGE_RAISED, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 2 && imageType == 0) || (hoverborderMethod == 2 && imageType == 1) || (pressedborderMethod == 2 && imageType == 2) )
		DrawEdge(bufferDC, &completeFillRect, EDGE_SUNKEN, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 3 && imageType == 0) || (hoverborderMethod == 3 && imageType == 1) || (pressedborderMethod == 3 && imageType == 2) )
		DrawEdge(bufferDC, &completeFillRect, EDGE_ETCHED, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 4 && imageType == 0) || (hoverborderMethod == 4 && imageType == 1) || (pressedborderMethod == 4 && imageType == 2) )
		DrawEdge(bufferDC, &completeFillRect, EDGE_BUMP, BF_RECT | BF_SOFT);

	BITMAP bm;
	GetObject(bufferBitmap, sizeof(BITMAP), &bm);

	BYTE *pPixel = (LPBYTE)bm.bmBits;

	int iSize = bm.bmHeight * bm.bmWidth;

	for (int j = 0; j < iSize; ++j)
	{
		//Remove Magic Pink
		if ( (pPixel[0] == 255 && pPixel[1] == 0 && pPixel[2] == 255) )
		{
			pPixel[0] = 0;
			pPixel[1] = 0;
			pPixel[2] = 0;
			pPixel[3] = 0;
		}
		else if ( !useAlphaMap )
		{
			pPixel[3] = 255;
		}

		pPixel+= 4;
	}

	BLENDFUNCTION blendFunction = { AC_SRC_OVER, 0, alphaValue, AC_SRC_ALPHA };

	AlphaBlend(hDC, x, y, width, height, bufferDC, 0, 0, width, height, blendFunction);

	bufferBitmap = (HBITMAP) SelectObject(bufferDC, oldBMP);
	DeleteObject(bufferBitmap);
	DeleteDC(bufferDC);
}

void IconTexture::CreateDCFromSource(HDC &destDC, HBITMAP &destBitmap, COLORREF _hueColor, int _hueIntensity, int _luminanceIntensity, int _saturationIntensity, COLORREF _mixColor, int _mixIntensity, int _iconSize, HICON noalphaMapIcon)
{
	//Paint Saturation and Hueing Effects in MemoryDC's
	destDC = CreateCompatibleDC(NULL);

	VOID *pvBits;
	BITMAPINFO alpha;
	if ( noalphaMapIcon != NULL && destBitmap == NULL )
	{
		ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
		alpha.bmiHeader.biWidth=_iconSize;      // Set size you need
		alpha.bmiHeader.biHeight=_iconSize;    // Set size you need
		alpha.bmiHeader.biPlanes=1;
		alpha.bmiHeader.biBitCount=32;      // Can be 8, 16, 32 bpp or other number
		alpha.bmiHeader.biSizeImage=0;
		alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
		alpha.bmiHeader.biClrUsed= 0;
		alpha.bmiHeader.biClrImportant= 0;;

		destBitmap = CreateDIBSection( destDC,
										&alpha,
										DIB_RGB_COLORS,
										&pvBits,
										NULL,
										0 );
	}

	SelectObject(destDC, destBitmap);

	HDC maskDC = NULL;
	HBITMAP maskBitmap = NULL;
	HBITMAP oldmaskBitmap = NULL;

	bool useMask = false;

	if ( noalphaMapIcon != NULL )
	{	
		maskDC = CreateCompatibleDC(destDC);
		maskBitmap = CreateDIBSection( maskDC,
										&alpha,
										DIB_RGB_COLORS,
										&pvBits,
										NULL,
										0 );

		oldmaskBitmap = (HBITMAP) SelectObject(maskDC, maskBitmap);

		DrawIconEx(maskDC, 0, 0, noalphaMapIcon, _iconSize, _iconSize, 0, NULL, DI_MASK);

		useMask = true;
	
		HBRUSH bgBrush = CreateSolidBrush(RGB(255,0,255));
		DrawIconEx(destDC, 0, 0, noalphaMapIcon, _iconSize, _iconSize, 0, bgBrush, DI_NORMAL);
		DeleteObject(bgBrush);
	}

	BITMAP bm;
	GetObject(destBitmap, sizeof(BITMAP), &bm);
	BYTE *pPixel = (LPBYTE)bm.bmBits;

	BITMAP bm_m;
	GetObject(maskBitmap, sizeof(BITMAP), &bm_m);
	BYTE *pPixel_m = (LPBYTE)bm_m.bmBits;

	BYTE r = 0;
	BYTE g = 0;
	BYTE b = 0;

	BYTE h = 0;
	BYTE s = 0;
	BYTE l = 0;
	
	BYTE hueh = 0;
	BYTE hues = 0;
	BYTE huel = 0;

	bool needHSLConversion = false;
	bool needMixColor = false;

	if ( _luminanceIntensity != 0 || _saturationIntensity != 100 )
		needHSLConversion = true;

	if ( _hueIntensity != 0 )
	{
		r = GetRValue(_hueColor);
		g = GetGValue(_hueColor);
		b = GetBValue(_hueColor);

		if ( r == b && b == g )
		{
			_hueIntensity = 0;
		}
		else
		{
			RGB2HSL(b, g, r, hueh, hues, huel); //Switch r and b due to AlphaMap/bm_b.bmHeight * bm_b.bmWidth
			needHSLConversion = true;
		}
	}

	if ( _mixIntensity != 0 )
	{
		_mixIntensity = (int) min(255, _mixIntensity*2.55);
		needMixColor = true;
	}

	int iSize = bm.bmHeight * bm.bmWidth;

	for (int j = 0; j < iSize; ++j)
	{
		//Magic Pink and PureTransparency
		if ( (pPixel[0] == 255 && pPixel[1] == 0 && pPixel[2] == 255) || (useMask && (pPixel_m[0] == 255 && pPixel_m[1] == 255 && pPixel_m[2] == 255)) )
		{
			pPixel[0] = 0;
			pPixel[1] = 0;
			pPixel[2] = 0;
			pPixel[3] = 0;

			transparencyNeeded = true;

			pPixel+= 4;
			if ( useMask )
				pPixel_m+= 4;
			continue;
		}

		if ( needHSLConversion || needMixColor )
		{	
			r = pPixel[0];
			g = pPixel[1];
			b = pPixel[2];

			if ( needHSLConversion )
			{
				RGB2HSL(r, g, b, h, s, l);

				if(_hueIntensity != 0)
					h = (BYTE) max(0, min(240, (h*(100-_hueIntensity)+hueh*_hueIntensity)/100));
			
				if(_luminanceIntensity != 0)
					l = (BYTE) max(0, min(240, ((l*100)+(120*_luminanceIntensity))/100));

				if(_saturationIntensity != 100)
				{
					if ( _saturationIntensity > 100 )
						s = (BYTE) max(0, min(240, (240*(_saturationIntensity-100))/100));				
					else	
						s = (BYTE) max(0, min(240, (s*_saturationIntensity)/100));
				}

				HSL2RGB(r, g, b, h, s, l);
			}

			if ( needMixColor )
			{
				// HueMix effect
				r = (BYTE)((r * (255 - _mixIntensity) + GetBValue(_mixColor) * _mixIntensity + 255) >> 8);
				g = (BYTE)((g * (255 - _mixIntensity) + GetGValue(_mixColor) * _mixIntensity + 255) >> 8);
				b = (BYTE)((b * (255 - _mixIntensity) + GetRValue(_mixColor) * _mixIntensity + 255) >> 8);
			}

			pPixel[0] = r;
			pPixel[1] = g;
			pPixel[2] = b;
		}

		if ( noalphaMapIcon != NULL )
			pPixel[3] = 255;

		pPixel+= 4;
		if ( useMask )
			pPixel_m+= 4;
	}

	if ( maskBitmap && maskDC )
	{
		maskBitmap = (HBITMAP) SelectObject(maskDC, oldmaskBitmap);
		DeleteObject(maskBitmap);
		DeleteDC(maskDC);
	}
}

void IconTexture::GetIconFromLnkTarget(CString path, int _iconSize, int imageMode, HICON &resultIcon)
{
	if ( !IsOS( OS_XP ) )
	{
		// Special case for link files
		// If the icon is set for the link, retrieve it
		SHFILEINFO shfi;
		SHGetFileInfo(path, 0, &shfi, sizeof(SHFILEINFO), SHGFI_ICONLOCATION);
		// Extract the icon
		if ( _PrivateExtractIcons != NULL )
			_PrivateExtractIcons(shfi.szDisplayName, shfi.iIcon, (_iconSize >= 32 ? 32 : 16), (_iconSize >= 32 ? 32 : 16), &resultIcon, NULL, 1, NULL);
	}

	// If the icon was not explicitly set in the linkfile, we'll use the icon from the target file
	if(resultIcon == NULL)
	{
		HRESULT hRes;				
		char szInfo[1024] = {0};
		IShellLink* psl;

		// Create the ShellLink object
		hRes = CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER, IID_IShellLink, (LPVOID*)&psl);
		if(SUCCEEDED(hRes))
		{
			IPersistFile* ppf;
			// Bind the ShellLink object to the Persistent File
			hRes = psl->QueryInterface(IID_IPersistFile, (LPVOID *) &ppf);
			if(SUCCEEDED(hRes))
			{
				WORD wsz[MAX_PATH];

				MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, path, -1, (LPWSTR)wsz, MAX_PATH);
				hRes = ppf->Load((LPCOLESTR)wsz, 0);
							
				if(SUCCEEDED(hRes))
				{
					// Retrieve the target of the link file
					psl->GetPath(szInfo, 1024, NULL, SLGP_UNCPRIORITY);
				}
			}
		}
		psl->Release();

		CIconExtractor m_IconExtractor;

		// Extract the icon from the link target
		if ( imageMode == 0 )
			resultIcon = m_IconExtractor.ExtractIconFromExeEx(szInfo, 0, memoryBitmap, _iconSize);
		else if (imageMode == 1 )
			resultIcon = m_IconExtractor.ExtractIconFromExeEx(szInfo, 0, memoryHoverBitmap, _iconSize);
		else if (imageMode == 2 )
			resultIcon = m_IconExtractor.ExtractIconFromExeEx(szInfo, 0, memoryPressedBitmap, _iconSize);
	}	
}
