/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2006 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "StdAfx.h"
#include "ImageTexture.h"

ImageTexture::ImageTexture() :
	memoryDC(0),
	memoryBitmap(0),

	memoryHoverDC(0),
	memoryHoverBitmap(0),

	memoryPressedDC(0),
	memoryPressedBitmap(0),

	memoryAnimDC(0),
	memoryAnimBitmap(0), 

	memoryHoverAnimDC(0),
	memoryHoverAnimBitmap(0), 

	imageWidth(-1),
	imageHeight(-1),
	hoverimageWidth(-1),
	hoverimageHeight(-1),
	pressedimageWidth(-1),
	pressedimageHeight(-1),
	
	frameWidth(-1),
	frameHeight(-1),
	hoverframeWidth(-1),
	hoverframeHeight(-1),

	transparencyNeeded(false),

	useAlphaMap(false),

	useHover(false),
	usePressed(false)
{
}

ImageTexture::~ImageTexture()
{
	if(memoryBitmap)
	{
		memoryBitmap = (HBITMAP) SelectObject(memoryDC, memoryBitmap);
		DeleteObject(memoryBitmap);
	}
	if(memoryDC)
		DeleteDC(memoryDC);

	if(memoryHoverBitmap)
	{
		memoryHoverBitmap = (HBITMAP) SelectObject(memoryHoverDC, memoryHoverBitmap);
		DeleteObject(memoryHoverBitmap);
	}
	if(memoryHoverDC)
		DeleteDC(memoryHoverDC);

	if(memoryPressedBitmap)
	{
		memoryPressedBitmap = (HBITMAP) SelectObject(memoryPressedDC, memoryPressedBitmap);
		DeleteObject(memoryPressedBitmap);
	}
	if(memoryPressedDC)
		DeleteDC(memoryPressedDC);

	if(memoryAnimBitmap)
	{
		memoryAnimBitmap = (HBITMAP) SelectObject(memoryAnimDC, memoryAnimBitmap);
		DeleteObject(memoryAnimBitmap);
	}
	if(memoryAnimDC)
		DeleteDC(memoryAnimDC);

	if(memoryHoverAnimBitmap)
	{
		memoryHoverAnimBitmap = (HBITMAP) SelectObject(memoryHoverAnimDC, memoryHoverAnimBitmap);
		DeleteObject(memoryHoverAnimBitmap);
	}
	if(memoryHoverAnimDC)
		DeleteDC(memoryHoverAnimDC);
}

void ImageTexture::configure(const string &uniqueKey, const string &groupKey, bool alphaMap)
{
	NameValuePair modeConstants[] =
	{
		{ "stretch", MULTIBLT_STRETCH },
		{ "tile", MULTIBLT_TILE },
		{ "tile-horizontal", MULTIBLT_TILEHORIZONTAL },
		{ "tile-vertical", MULTIBLT_TILEVERTICAL },
		{ "tilehorizontal", MULTIBLT_TILEHORIZONTAL },
		{ "tilevertical", MULTIBLT_TILEVERTICAL },
		{ 0, 0 }
	};

	NameValuePair borderValues[] = 
	{
		{ ".none", 0 },
		{ "none", 0 },
		{ "raised", 1 },
		{ "sunken", 2 },
		{ "etched", 3 },
		{ "bump", 4 },
		{ 0, 0 }
	};

	useAlphaMap = alphaMap;

	//Always Needed Default Values
	//Normal (for Hover) and Hover (for Pressed)
	//--------------------------------------------
	hueColor = hoverhueColor = Getx2RCColor(uniqueKey, groupKey, "HueColor", RGB(255, 255, 255));
	hueIntensity = hoverhueIntensity = Getx2RCInt(uniqueKey, groupKey, "HueIntensity", 0, 0, 100);
	luminanceIntensity = hoverluminanceIntensity = Getx2RCInt(uniqueKey, groupKey, "LuminanceIntensity", 0, -100, 100);
	saturationIntensity = hoversaturationIntensity = Getx2RCInt(uniqueKey, groupKey, "SaturationIntensity", 100, 0, 200);

	mixColor = hovermixColor = Getx2RCColor(uniqueKey, groupKey, "MixColor", RGB(255, 255, 255));
	mixIntensity = hovermixIntensity = Getx2RCInt(uniqueKey, groupKey, "MixIntensity", 0, 0, 100);

	leftEdge = hoverleftEdge = Getx2RCInt(uniqueKey, groupKey, "ImageLeftEdge", 0, 0);
	topEdge = hovertopEdge = Getx2RCInt(uniqueKey, groupKey, "ImageTopEdge", 0, 0);
	rightEdge = hoverrightEdge = Getx2RCInt(uniqueKey, groupKey, "ImageRightEdge", 0, 0);
	bottomEdge = hoverbottomEdge = Getx2RCInt(uniqueKey, groupKey, "ImageBottomEdge", 0, 0);

	mode = hovermode = Getx2RCNamedValue(uniqueKey, groupKey, "ImageMode", modeConstants, MULTIBLT_STRETCH);
	
	borderMethod = hoverborderMethod = Getx2RCNamedValue(uniqueKey, groupKey, "BorderMethod", borderValues, 0);
	//--------------------------------------------

	string imageFile = Getx2RCString(uniqueKey, groupKey, "Image", "");

	if ( !imageFile.empty() )
	{
		memoryBitmap = LoadLSImage(imageFile.c_str(), 0);

		if ( memoryBitmap != NULL )
		{
			GetLSBitmapSize(memoryBitmap, &imageWidth, &imageHeight);

			memoryDC = CreateCompatibleDC(NULL);
			SelectObject(memoryDC, memoryBitmap);

			CString strMessage = "";
			if ( imageWidth != -1 && leftEdge+rightEdge >= imageWidth )
			{
				strMessage.Format("Label: %s\nPlease make sure that the Image Width (%i) is greater than the sum of ImageLeftEdge (%i) and ImageRightEdge (%i)", uniqueKey.c_str(), imageWidth, leftEdge, rightEdge);
				MessageBox(NULL, strMessage, "Configuration Mistake", MB_ICONEXCLAMATION | MB_OK);
			}
			if ( imageHeight != -1 && topEdge+bottomEdge >= imageHeight )
			{
				strMessage.Format("Label: %s\nPlease make sure that the Image Height (%i) is greater than the sum of ImageTopEdge (%i) and ImageBottomEdge (%i)", uniqueKey.c_str(), imageHeight, topEdge, bottomEdge);
				MessageBox(NULL, strMessage, "Configuration Mistake", MB_ICONEXCLAMATION | MB_OK);
			}

			if (useAlphaMap)
			{
				BITMAP bm;
				GetObject(memoryBitmap, sizeof(BITMAP), &bm);

				if ((bm.bmBits != 0) && (bm.bmBitsPixel == 32))
					PrepareAlphaChannel(memoryBitmap, 0);
			}
			else if (memoryDC && memoryBitmap)
			{
				paintEffects(memoryDC, imageWidth, imageHeight, 0);
			}
		}
		else
		{
			transparencyNeeded = true;
		}
	}
	else
	{
		transparencyNeeded = true;
	}

	imageFile = Getx2RCString(uniqueKey, groupKey, "HoverImage", "");

	if (!imageFile.empty())
	{
		memoryHoverBitmap = LoadLSImage(imageFile.c_str(), 0);
		
		if ( memoryHoverBitmap != NULL )
		{	
			useHover = true;

			hoverhueColor = Getx2RCColor(uniqueKey, groupKey, "HoverHueColor", hueColor);
			hoverhueIntensity = Getx2RCInt(uniqueKey, groupKey, "HoverHueIntensity", hueIntensity, 0, 100);
			hoverluminanceIntensity = Getx2RCInt(uniqueKey, groupKey, "HoverLuminanceIntensity", luminanceIntensity, -100, 100);
			hoversaturationIntensity = Getx2RCInt(uniqueKey, groupKey, "HoverSaturationIntensity", saturationIntensity, 0, 200);

			hovermixColor = Getx2RCColor(uniqueKey, groupKey, "HoverMixColor", mixColor);
			hovermixIntensity = Getx2RCInt(uniqueKey, groupKey, "HoverMixIntensity", mixIntensity, 0, 100);

			hoverleftEdge = Getx2RCInt(uniqueKey, groupKey, "HoverImageLeftEdge", leftEdge, 0);
			hovertopEdge = Getx2RCInt(uniqueKey, groupKey, "HoverImageTopEdge", topEdge, 0);
			hoverrightEdge = Getx2RCInt(uniqueKey, groupKey, "HoverImageRightEdge", rightEdge, 0);
			hoverbottomEdge = Getx2RCInt(uniqueKey, groupKey, "HoverImageBottomEdge", bottomEdge, 0);

			hovermode = Getx2RCNamedValue(uniqueKey, groupKey, "HoverImageMode", modeConstants, mode);
	
			hoverborderMethod = Getx2RCNamedValue(uniqueKey, groupKey, "HoverBorderMethod", borderValues, borderMethod);

			GetLSBitmapSize(memoryHoverBitmap, &hoverimageWidth, &hoverimageHeight);

			memoryHoverDC = CreateCompatibleDC(NULL);
			SelectObject(memoryHoverDC, memoryHoverBitmap);

			CString strMessage = "";
			if ( hoverimageWidth != -1 && hoverleftEdge+hoverrightEdge >= hoverimageWidth )
			{
				strMessage.Format("Label: %s\nPlease make sure that the HoverImage Width (%i) is greater than the sum of HoverImageLeftEdge (%i) and HoverImageRightEdge (%i)", uniqueKey.c_str(), hoverimageWidth, hoverleftEdge, hoverrightEdge);
				MessageBox(NULL, strMessage, "Configuration Mistake", MB_ICONEXCLAMATION | MB_OK);
			}
			if ( hoverimageHeight != -1 && hovertopEdge+hoverbottomEdge >= hoverimageHeight )
			{
				strMessage.Format("Label: %s\nPlease make sure that the HoverImage Height (%i) is greater than the sum of HoverImageTopEdge (%i) and HoverImageBottomEdge (%i)", uniqueKey.c_str(), hoverimageHeight, hovertopEdge, hoverbottomEdge);
				MessageBox(NULL, strMessage, "Configuration Mistake", MB_ICONEXCLAMATION | MB_OK);
			}

			if (useAlphaMap)
			{
				BITMAP bm;
				GetObject(memoryHoverBitmap, sizeof(BITMAP), &bm);

				if ((bm.bmBits != 0) && (bm.bmBitsPixel == 32))
					PrepareAlphaChannel(memoryHoverBitmap, 1);
			}
			else if (memoryHoverDC && memoryHoverBitmap)
			{
				paintEffects(memoryHoverDC, hoverimageWidth, hoverimageHeight, 1);
			}
		}
	}

	imageFile = Getx2RCString(uniqueKey, groupKey, "PressedImage", "");

	if (!imageFile.empty())
	{
		memoryPressedBitmap = LoadLSImage(imageFile.c_str(), 0);
		
		if ( memoryPressedBitmap != NULL )
		{	
			usePressed = true;

			pressedhueColor = Getx2RCColor(uniqueKey, groupKey, "PressedHueColor", hoverhueColor);
			pressedhueIntensity = Getx2RCInt(uniqueKey, groupKey, "PressedHueIntensity", hoverhueIntensity, 0, 100);
			pressedluminanceIntensity = Getx2RCInt(uniqueKey, groupKey, "PressedLuminanceIntensity", hoverluminanceIntensity, -100, 100);
			pressedsaturationIntensity = Getx2RCInt(uniqueKey, groupKey, "PressedSaturationIntensity", hoversaturationIntensity, 0, 200);

			pressedmixColor = Getx2RCColor(uniqueKey, groupKey, "PressedMixColor", hovermixColor);
			pressedmixIntensity = Getx2RCInt(uniqueKey, groupKey, "PressedMixIntensity", hovermixIntensity, 0, 100);
	
			pressedleftEdge = Getx2RCInt(uniqueKey, groupKey, "PressedImageLeftEdge", hoverleftEdge, 0);
			pressedtopEdge = Getx2RCInt(uniqueKey, groupKey, "PressedImageTopEdge", hovertopEdge, 0);
			pressedrightEdge = Getx2RCInt(uniqueKey, groupKey, "PressedImageRightEdge", hoverrightEdge, 0);
			pressedbottomEdge = Getx2RCInt(uniqueKey, groupKey, "PressedImageBottomEdge", hoverbottomEdge, 0);
	
			pressedmode = Getx2RCNamedValue(uniqueKey, groupKey, "PressedImageMode", modeConstants, hovermode);
	
			pressedborderMethod = Getx2RCNamedValue(uniqueKey, groupKey, "PressedBorderMethod", borderValues, hoverborderMethod);

			GetLSBitmapSize(memoryPressedBitmap, &pressedimageWidth, &pressedimageHeight);

			memoryPressedDC = CreateCompatibleDC(NULL);
			SelectObject(memoryPressedDC, memoryPressedBitmap);

			CString strMessage = "";
			if ( pressedimageWidth != -1 && pressedleftEdge+pressedrightEdge >= pressedimageWidth )
			{
				strMessage.Format("Label: %s\nPlease make sure that the PressedImage Width (%i) is greater than the sum of PressedImageLeftEdge (%i) and PressedImageRightEdge (%i)", uniqueKey.c_str(), pressedimageWidth, pressedleftEdge, pressedrightEdge);
				MessageBox(NULL, strMessage, "Configuration Mistake", MB_ICONEXCLAMATION | MB_OK);
			}
			if ( pressedimageHeight != -1 && pressedtopEdge+pressedbottomEdge >= pressedimageHeight )
			{
				strMessage.Format("Label: %s\nPlease make sure that the PressedImage Height (%i) is greater than the sum of PressedImageTopEdge (%i) and PressedImageBottomEdge (%i)", uniqueKey.c_str(), pressedimageHeight, pressedtopEdge, pressedbottomEdge);
				MessageBox(NULL, strMessage, "Configuration Mistake", MB_ICONEXCLAMATION | MB_OK);
			}

			if (useAlphaMap)
			{
				BITMAP bm;
				GetObject(memoryPressedBitmap, sizeof(BITMAP), &bm);

				if ((bm.bmBits != 0) && (bm.bmBitsPixel == 32))
					PrepareAlphaChannel(memoryPressedBitmap, 2);
			}
			else if (memoryPressedDC && memoryPressedBitmap)
			{
				paintEffects(memoryPressedDC, pressedimageWidth, pressedimageHeight, 2);
			}
		}
	}
	
	imageFile = GetxRCString(uniqueKey, "Frames", "");

	int animimageWidth = 0;
	int animimageHeight = 0;

	//Check for Single Image animations
	if ( !imageFile.empty() )
	{
		memoryAnimBitmap = LoadLSImage(imageFile.c_str(), 0);

		if ( memoryAnimBitmap != NULL )
		{
			GetLSBitmapSize(memoryAnimBitmap, &animimageWidth, &animimageHeight);

			memoryAnimDC = CreateCompatibleDC(NULL);
			SelectObject(memoryAnimDC, memoryAnimBitmap);

			frameWidth = GetxRCInt(uniqueKey, "FrameWidth", 0);
			frameHeight = GetxRCInt(uniqueKey, "FrameHeight", 0);
		}
	}
	//Check for Multi Image animations
	else
	{
		StringList animation = GetxRCLineList( ("*" + uniqueKey + "Frames") );
	
		if ( !animation.empty() )
		{
			frameWidth = GetxRCInt(uniqueKey, "FrameWidth", 0);
			frameHeight = GetxRCInt(uniqueKey, "FrameHeight", 0);

			animimageWidth = frameWidth*animation.size();
			animimageHeight = frameHeight*animation.size();

			memoryAnimDC = CreateCompatibleDC(NULL);

			BITMAPINFO alpha;
			ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
			alpha.bmiHeader.biWidth=animimageWidth;      // Set size you need
			alpha.bmiHeader.biHeight=animimageHeight;    // Set size you need
			alpha.bmiHeader.biPlanes=1;
			alpha.bmiHeader.biBitCount=32;      // Can be 8, 16, 32 bpp or other number
			alpha.bmiHeader.biSizeImage=0;
			alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
			alpha.bmiHeader.biClrUsed= 0;
			alpha.bmiHeader.biClrImportant= 0;
			VOID *pvBits;

			memoryAnimBitmap = CreateDIBSection( memoryAnimDC,
													&alpha,
													DIB_RGB_COLORS,
													&pvBits,
													NULL,
													0 );

			SelectObject(memoryAnimDC, memoryAnimBitmap);

			HBITMAP tmpBitmap = NULL;
			HDC tmpDC = NULL;

			int x = 0;

			for (StringListIterator sit = animation.begin(); sit != animation.end(); sit++)
			{
				tmpBitmap = LoadLSImage((*sit).c_str(), 0);

				tmpDC = CreateCompatibleDC(NULL);
				SelectObject(tmpDC, tmpBitmap);

				BitBlt(memoryAnimDC, x, 0, frameWidth, frameHeight, tmpDC, 0, 0, SRCCOPY);

				x = x+frameWidth;

				if(tmpBitmap)
				{
					tmpBitmap = (HBITMAP) SelectObject(tmpDC, tmpBitmap);
					DeleteObject(tmpBitmap);
				}
				if(tmpDC)
					DeleteDC(tmpDC);
			}
		}
	}

	if ( memoryAnimBitmap != NULL && memoryAnimDC != NULL )
	{
		if (useAlphaMap)
		{
			BITMAP bm;
			GetObject(memoryAnimBitmap, sizeof(BITMAP), &bm);

			if ((bm.bmBits != 0) && (bm.bmBitsPixel == 32))
				PrepareAlphaChannel(memoryAnimBitmap, 0);
		}
		else if (memoryAnimDC && memoryAnimBitmap)
		{
			paintEffects(memoryAnimDC, animimageWidth, animimageHeight, 0);
		}
	}
	
	imageFile = GetxRCString(uniqueKey, "HoverFrames", "");
	
	int hoveranimimageWidth = 0;
	int hoveranimimageHeight = 0;

	if (!imageFile.empty())
	{
		memoryHoverAnimBitmap = LoadLSImage( imageFile.c_str(), 0 );

		if ( memoryHoverAnimBitmap != NULL )
		{
			useHover = true;

			GetLSBitmapSize(memoryHoverAnimBitmap, &hoveranimimageWidth, &hoveranimimageHeight);
				
			memoryHoverAnimDC = CreateCompatibleDC(NULL);
			SelectObject(memoryHoverAnimDC, memoryHoverAnimBitmap);

			hoverframeWidth = GetxRCInt(uniqueKey, "HoverFrameWidth", 0);
			hoverframeHeight = GetxRCInt(uniqueKey, "HoverFrameHeight", 0);
		}
	}
	else
	{
		StringList animation = GetxRCLineList( ("*" + uniqueKey + "HoverFrames") );
	
		if ( !animation.empty() )
		{
			hoverframeWidth = GetxRCInt(uniqueKey, "HoverFrameWidth", 0);
			hoverframeHeight = GetxRCInt(uniqueKey, "HoverFrameHeight", 0);

			hoveranimimageWidth = hoverframeWidth*animation.size();
			hoveranimimageHeight = hoverframeHeight*animation.size();

			memoryHoverAnimDC = CreateCompatibleDC(NULL);

			BITMAPINFO alpha;
			ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
			alpha.bmiHeader.biWidth=hoveranimimageWidth;      // Set size you need
			alpha.bmiHeader.biHeight=hoveranimimageHeight;    // Set size you need
			alpha.bmiHeader.biPlanes=1;
			alpha.bmiHeader.biBitCount=32;      // Can be 8, 16, 32 bpp or other number
			alpha.bmiHeader.biSizeImage=0;
			alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
			alpha.bmiHeader.biClrUsed= 0;
			alpha.bmiHeader.biClrImportant= 0;
			VOID *pvBits;

			memoryHoverAnimBitmap = CreateDIBSection( memoryHoverAnimDC,
													&alpha,
													DIB_RGB_COLORS,
													&pvBits,
													NULL,
													0 );

			SelectObject(memoryHoverAnimDC, memoryHoverAnimBitmap);

			HBITMAP tmpBitmap = NULL;
			HDC tmpDC = NULL;

			int x = 0;

			for (StringListIterator sit = animation.begin(); sit != animation.end(); sit++)
			{
				tmpBitmap = LoadLSImage((*sit).c_str(), 0);

				tmpDC = CreateCompatibleDC(NULL);
				SelectObject(tmpDC, tmpBitmap);

				BitBlt(memoryHoverAnimDC, x, 0, hoverframeWidth, hoverframeHeight, tmpDC, 0, 0, SRCCOPY);

				x = x+hoverframeWidth;

				if(tmpBitmap)
				{
					tmpBitmap = (HBITMAP) SelectObject(tmpDC, tmpBitmap);
					DeleteObject(tmpBitmap);
				}
				if(tmpDC)
					DeleteDC(tmpDC);
			}
		}
	}
	
	if ( memoryHoverAnimBitmap != NULL && memoryHoverAnimDC != NULL )
	{
		if (useAlphaMap)
		{
			BITMAP bm;
			GetObject(memoryHoverAnimBitmap, sizeof(BITMAP), &bm);

			if ((bm.bmBits != 0) && (bm.bmBitsPixel == 32))
				PrepareAlphaChannel(memoryHoverAnimBitmap, 1);
		}
		else if (memoryHoverAnimDC && memoryHoverAnimBitmap)
		{
			paintEffects(memoryHoverAnimDC, hoveranimimageWidth, hoveranimimageHeight, 1);
		}
	}
}

void ImageTexture::apply(HDC hDC, int x, int y, int width, int height, int transparencyMode, int imageType)
{
	if ( width <= 0 || height <= 0 )
		return;

	if ( imageType == 2 && memoryPressedBitmap == 0 )
		imageType = 1;

	if ( imageType == 1 && memoryHoverBitmap == 0 )
		imageType = 0;

	if ( imageType == 0 && memoryBitmap == 0 )
		return;

	if (transparencyMode == 2 || transparencyMode == 3)
	{
		HDC bufferDC = CreateCompatibleDC(hDC);
		HBITMAP bufferBitmap = CreateCompatibleBitmap(hDC, width, height);
		HBITMAP oldBMP = (HBITMAP) SelectObject(bufferDC, bufferBitmap);

		SetStretchBltMode(bufferDC, STRETCH_DELETESCANS);

		if (imageType == 0)
		{
			MultiBlt(bufferDC, 0, 0, width, height,
						memoryDC, 0, 0, imageWidth, imageHeight,
						leftEdge, topEdge, rightEdge, bottomEdge,
						mode, SRCCOPY);
		}
		else if (imageType == 1)
		{
			MultiBlt(bufferDC, 0, 0, width, height,
						memoryHoverDC, 0, 0, hoverimageWidth, hoverimageHeight,
						hoverleftEdge, hovertopEdge, hoverrightEdge, hoverbottomEdge,
						hovermode, SRCCOPY);
		}
		else if (imageType == 2)
		{
			MultiBlt(bufferDC, 0, 0, width, height,
						memoryPressedDC, 0, 0, pressedimageWidth, pressedimageHeight,
						pressedleftEdge, pressedtopEdge, pressedrightEdge, pressedbottomEdge,
						pressedmode, SRCCOPY);
		}
			
		TransparentBltLS(hDC, x, y, width, height,
							bufferDC, 0, 0,
							RGB(255, 0, 255));

		bufferBitmap = (HBITMAP) SelectObject(bufferDC, oldBMP);
		DeleteObject(bufferBitmap);
		DeleteDC(bufferDC);
	}
	else
	{
		SetStretchBltMode(hDC, STRETCH_DELETESCANS);

		if (imageType == 0)
		{
			MultiBlt(hDC, x, y, width, height,
						memoryDC, 0, 0, imageWidth, imageHeight,
						leftEdge, topEdge, rightEdge, bottomEdge,
						mode, SRCCOPY);
		}
		else if (imageType == 1)
		{
			MultiBlt(hDC, x, y, width, height,
						memoryHoverDC, 0, 0, hoverimageWidth, hoverimageHeight,
						hoverleftEdge, hovertopEdge, hoverrightEdge, hoverbottomEdge,
						hovermode, SRCCOPY);
		}
		else if (imageType == 2)
		{
			MultiBlt(hDC, x, y, width, height,
						memoryPressedDC, 0, 0, pressedimageWidth, pressedimageHeight,
						pressedleftEdge, pressedtopEdge, pressedrightEdge, pressedbottomEdge,
						pressedmode, SRCCOPY);
		}
	}

	RECT completeFillRect = {x, y, x+width, y+height};

	if ( (borderMethod == 1 && imageType == 0) || (hoverborderMethod == 1 && imageType == 1) || (pressedborderMethod == 1 && imageType == 2) )
		DrawEdge(hDC, &completeFillRect, EDGE_RAISED, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 2 && imageType == 0) || (hoverborderMethod == 2 && imageType == 1) || (pressedborderMethod == 2 && imageType == 2) )
		DrawEdge(hDC, &completeFillRect, EDGE_SUNKEN, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 3 && imageType == 0) || (hoverborderMethod == 3 && imageType == 1) || (pressedborderMethod == 3 && imageType == 2) )
		DrawEdge(hDC, &completeFillRect, EDGE_ETCHED, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 4 && imageType == 0) || (hoverborderMethod == 4 && imageType == 1) || (pressedborderMethod == 4 && imageType == 2) )
		DrawEdge(hDC, &completeFillRect, EDGE_BUMP, BF_RECT | BF_SOFT);
}

void ImageTexture::applyOverlay(HDC hDC, int x, int y, int width, int height, int alphaValue, int imageType)
{
	if ( width <= 0 || height <= 0 )
		return;

	if ( imageType == 2 && memoryPressedBitmap == 0 )
		imageType = 1;

	if ( imageType == 1 && memoryHoverBitmap == 0 )
		imageType = 0;

	if ( imageType == 0 && memoryBitmap == 0 )
		return;

	HDC bufferDC = CreateCompatibleDC(NULL);

	BITMAPINFO alpha;
	ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
	alpha.bmiHeader.biWidth=width;      // Set size you need
	alpha.bmiHeader.biHeight=height;    // Set size you need
	alpha.bmiHeader.biPlanes=1;
	alpha.bmiHeader.biBitCount=32;      // Can be 8, 16, 32 bpp or other number
	alpha.bmiHeader.biSizeImage=0;
	alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	alpha.bmiHeader.biClrUsed= 0;
	alpha.bmiHeader.biClrImportant= 0;
	VOID *pvBits;

	HBITMAP bufferBitmap = CreateDIBSection( bufferDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

	HBITMAP oldBMP = (HBITMAP) SelectObject(bufferDC, bufferBitmap);

	SetStretchBltMode(bufferDC, STRETCH_DELETESCANS);

	if (imageType == 0)
	{
		MultiBlt(bufferDC, 0, 0, width, height,
					memoryDC, 0, 0, imageWidth, imageHeight,
					leftEdge, topEdge, rightEdge, bottomEdge,
					mode, SRCCOPY);
	}
	else if (imageType == 1)
	{
		MultiBlt(bufferDC, 0, 0, width, height,
					memoryHoverDC, 0, 0, hoverimageWidth, hoverimageHeight,
					hoverleftEdge, hovertopEdge, hoverrightEdge, hoverbottomEdge,
					hovermode, SRCCOPY);
	}
	else if (imageType == 2)
	{
		MultiBlt(bufferDC, 0, 0, width, height,
					memoryPressedDC, 0, 0, pressedimageWidth, pressedimageHeight,
					pressedleftEdge, pressedtopEdge, pressedrightEdge, pressedbottomEdge,
					pressedmode, SRCCOPY);
	}

	RECT completeFillRect = {0, 0, width, height};

	if ( (borderMethod == 1 && imageType == 0) || (hoverborderMethod == 1 && imageType == 1) || (pressedborderMethod == 1 && imageType == 2) )
		DrawEdge(bufferDC, &completeFillRect, EDGE_RAISED, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 2 && imageType == 0) || (hoverborderMethod == 2 && imageType == 1) || (pressedborderMethod == 2 && imageType == 2) )
		DrawEdge(bufferDC, &completeFillRect, EDGE_SUNKEN, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 3 && imageType == 0) || (hoverborderMethod == 3 && imageType == 1) || (pressedborderMethod == 3 && imageType == 2) )
		DrawEdge(bufferDC, &completeFillRect, EDGE_ETCHED, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 4 && imageType == 0) || (hoverborderMethod == 4 && imageType == 1) || (pressedborderMethod == 4 && imageType == 2) )
		DrawEdge(bufferDC, &completeFillRect, EDGE_BUMP, BF_RECT | BF_SOFT);

	BITMAP bm;
	GetObject(bufferBitmap, sizeof(BITMAP), &bm);

	BYTE *pPixel = (LPBYTE)bm.bmBits;

	int iSize = bm.bmHeight * bm.bmWidth;

	for (int j = 0; j < iSize; ++j)
	{
		//Remove Magic Pink
		if ( (pPixel[0] == 255 && pPixel[1] == 0 && pPixel[2] == 255) )
		{
			pPixel[0] = 0;
			pPixel[1] = 0;
			pPixel[2] = 0;
			pPixel[3] = 0;
		}
		else if ( !useAlphaMap || pPixel[3] == 0 )
		{
			pPixel[3] = 255;
		}

		pPixel+= 4;
	}
	
	BLENDFUNCTION blendFunction = { AC_SRC_OVER, 0, alphaValue, AC_SRC_ALPHA };

	AlphaBlend(hDC, x, y, width, height, bufferDC, 0, 0, width, height, blendFunction);

	bufferBitmap = (HBITMAP) SelectObject(bufferDC, oldBMP);
	DeleteObject(bufferBitmap);
	DeleteDC(bufferDC);
}

void ImageTexture::applyAnim(HDC hDC, int x, int y, int width, int height, int transparencyMode, int imageType, int frameCounter)
{
	if ( width <= 0 || height <= 0 )
		return;

	if (imageType == 1 && memoryHoverAnimBitmap == 0)
		imageType = 0;
	
	if (imageType == 0 && memoryAnimBitmap == 0)
		return;

	if (transparencyMode == 2 || transparencyMode == 3)
	{
		HDC bufferDC = CreateCompatibleDC(hDC);
		HBITMAP bufferBitmap = CreateCompatibleBitmap(hDC, width, height);
		HBITMAP oldBMP = (HBITMAP) SelectObject(bufferDC, bufferBitmap);

		SetStretchBltMode(bufferDC, STRETCH_DELETESCANS);

		if (imageType == 0)
		{
			MultiBlt(bufferDC, 0, 0, width, height,
				memoryAnimDC, (frameCounter-1)*frameWidth, 0, frameWidth, frameHeight,
				leftEdge, topEdge, rightEdge, bottomEdge,
				mode, SRCCOPY);
		}
		else if (imageType == 1)
		{
			MultiBlt(bufferDC, 0, 0, width, height,
				memoryHoverAnimDC, (frameCounter-1)*hoverframeWidth, 0, hoverframeWidth, hoverframeHeight,
				hoverleftEdge, hovertopEdge, hoverrightEdge, hoverbottomEdge,
				hovermode, SRCCOPY);
		}

		TransparentBltLS(hDC, x, y, width, height,
			bufferDC, 0, 0,
			RGB(255, 0, 255));

		bufferBitmap = (HBITMAP) SelectObject(bufferDC, oldBMP);
		DeleteObject(bufferBitmap);
		DeleteDC(bufferDC);
	}
	else
	{
		SetStretchBltMode(hDC, STRETCH_DELETESCANS);

		if (imageType == 0)
		{
			MultiBlt(hDC, x, y, width, height,
					memoryAnimDC, (frameCounter-1)*frameWidth, 0, frameWidth, frameHeight,
					leftEdge, topEdge, rightEdge, bottomEdge,
					mode, SRCCOPY);
		}
		else if (imageType == 1)
		{
			MultiBlt(hDC, x, y, width, height,
					memoryHoverAnimDC, (frameCounter-1)*hoverframeWidth, 0, hoverframeWidth, hoverframeHeight,
					hoverleftEdge, hovertopEdge, hoverrightEdge, hoverbottomEdge,
					hovermode, SRCCOPY);
		}
	}

	RECT completeFillRect = {x, y, x+width, y+height};

	if ( (borderMethod == 1 && imageType == 0) || (hoverborderMethod == 1 && imageType == 1) || (pressedborderMethod == 1 && imageType == 2) )
		DrawEdge(hDC, &completeFillRect, EDGE_RAISED, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 2 && imageType == 0) || (hoverborderMethod == 2 && imageType == 1) || (pressedborderMethod == 2 && imageType == 2) )
		DrawEdge(hDC, &completeFillRect, EDGE_SUNKEN, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 3 && imageType == 0) || (hoverborderMethod == 3 && imageType == 1) || (pressedborderMethod == 3 && imageType == 2) )
		DrawEdge(hDC, &completeFillRect, EDGE_ETCHED, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 4 && imageType == 0) || (hoverborderMethod == 4 && imageType == 1) || (pressedborderMethod == 4 && imageType == 2) )
		DrawEdge(hDC, &completeFillRect, EDGE_BUMP, BF_RECT | BF_SOFT);
}

void ImageTexture::PrepareAlphaChannel(HBITMAP &bmp, int imageType)
{
	//Paint Saturation and Hueing Effects in MemoryDC's
	COLORREF _hueColor = RGB(255,255,255);
	int _hueIntensity = 0;
	int _luminanceIntensity = 0;
	int _saturationIntensity = 100;

	COLORREF _mixColor = RGB(255,255,255);
	int _mixIntensity = 0;

	if (imageType == 0)
	{
		_hueColor = hueColor;
		_hueIntensity = hueIntensity;
		_luminanceIntensity = luminanceIntensity;
		_saturationIntensity = saturationIntensity;

		_mixColor = mixColor;
		_mixIntensity = mixIntensity;
	}
	else if (imageType == 1)
	{
		_hueColor = hoverhueColor;
		_hueIntensity = hoverhueIntensity;
		_luminanceIntensity = hoverluminanceIntensity;
		_saturationIntensity = hoversaturationIntensity;

		_mixColor = hovermixColor;
		_mixIntensity = hovermixIntensity;
	}
	else if (imageType == 2)
	{
		_hueColor = pressedhueColor;
		_hueIntensity = pressedhueIntensity;
		_luminanceIntensity = pressedluminanceIntensity;
		_saturationIntensity = pressedsaturationIntensity;

		_mixColor = pressedmixColor;
		_mixIntensity = pressedmixIntensity;
	}

	BITMAP bm;
	GetObject(bmp, sizeof(BITMAP), &bm);

	BYTE *pPixel = (LPBYTE)bm.bmBits;

	BYTE r = 0;
	BYTE g = 0;
	BYTE b = 0;

	BYTE h = 0;
	BYTE s = 0;
	BYTE l = 0;
	
	BYTE hueh = 0;
	BYTE hues = 0;
	BYTE huel = 0;

	bool needHSLConversion = false;
	bool needMixColor = false;

	if ( _luminanceIntensity != 0 || _saturationIntensity != 100 )
		needHSLConversion = true;

	if ( _hueIntensity != 0 )
	{
		r = GetRValue(_hueColor);
		g = GetGValue(_hueColor);
		b = GetBValue(_hueColor);

		if ( r == b && b == g )
		{
			_hueIntensity = 0;
		}
		else
		{
			RGB2HSL(b, g, r, hueh, hues, huel); //Switch r and b due to AlphaMap/bm_b.bmHeight * bm_b.bmWidth
			needHSLConversion = true;
		}
	}

	if ( _mixIntensity != 0 )
	{
		_mixIntensity = (int) min(255, _mixIntensity*2.55);
		needMixColor = true;
	}

	int iSize = bm.bmHeight * bm.bmWidth;

	for (int i = 0; i < iSize; ++i)
	{
		//Magic Pink
		if (pPixel[0] == 255 && pPixel[1] == 0 && pPixel[2] == 255)
			pPixel[3] = 0;

		//Prepare Text
		if (pPixel[3] == 0)
			pPixel[3] = 1;


		r = pPixel[0];
		g = pPixel[1];
		b = pPixel[2];

		if ( needHSLConversion )
		{
			//AlphaMap Colorizing
			RGB2HSL(r, g, b, h, s, l);

			if(_hueIntensity != 0)
				h = (BYTE) max(0, min(240, (h*(100-_hueIntensity)+hueh*_hueIntensity)/100));
		
			if(_luminanceIntensity != 0)
				l = (BYTE) max(0, min(240, ((l*100)+(120*_luminanceIntensity))/100));

			if(_saturationIntensity != 100)
			{
				if ( _saturationIntensity > 100 )
					s = (BYTE) max(0, min(240, (240*(_saturationIntensity-100))/100));				
				else
					s = (BYTE) max(0, min(240, (s*_saturationIntensity)/100));
			}

			HSL2RGB(r, g, b, h, s, l);
		}

		if ( needMixColor )
		{
			// HueMix effect
			r = (BYTE)((r * (255 - _mixIntensity) + GetBValue(_mixColor) * _mixIntensity + 255) >> 8);
			g = (BYTE)((g * (255 - _mixIntensity) + GetGValue(_mixColor) * _mixIntensity + 255) >> 8);
			b = (BYTE)((b * (255 - _mixIntensity) + GetRValue(_mixColor) * _mixIntensity + 255) >> 8);
		}
		
		//PreMultiply
		if ( pPixel[3] == 1 )
		{
			pPixel[0] = 0;
			pPixel[1] = 0;
			pPixel[2] = 0;
		}
		else
		{
			pPixel[0] = r * pPixel[3] / 255;
			pPixel[1] = g * pPixel[3] / 255;
			pPixel[2] = b * pPixel[3] / 255;
		}
	
		pPixel+= 4;
	}
}

void ImageTexture::paintEffects(HDC &destHDC, int width, int height, int imageType)
{
	//Paint Saturation and Hueing Effects in MemoryDC's
	COLORREF _hueColor = RGB(255,255,255);
	int _hueIntensity = 0;
	int _luminanceIntensity = 0;
	int _saturationIntensity = 100;

	COLORREF _mixColor = RGB(255,255,255);
	int _mixIntensity = 0;

	if (imageType == 0)
	{
		_hueColor = hueColor;
		_hueIntensity = hueIntensity;
		_luminanceIntensity = luminanceIntensity;
		_saturationIntensity = saturationIntensity;

		_mixColor = mixColor;
		_mixIntensity = mixIntensity;
	}
	else if (imageType == 1)
	{
		_hueColor = hoverhueColor;
		_hueIntensity = hoverhueIntensity;
		_luminanceIntensity = hoverluminanceIntensity;
		_saturationIntensity = hoversaturationIntensity;

		_mixColor = hovermixColor;
		_mixIntensity = hovermixIntensity;
	}
	else if (imageType == 2)
	{
		_hueColor = pressedhueColor;
		_hueIntensity = pressedhueIntensity;
		_luminanceIntensity = pressedluminanceIntensity;
		_saturationIntensity = pressedsaturationIntensity;

		_mixColor = pressedmixColor;
		_mixIntensity = pressedmixIntensity;
	}

	int iy = 0;
	int ix = 0;
	COLORREF cl = 0;
	BYTE r = 0;
	BYTE g = 0;
	BYTE b = 0;

	BYTE h = 0;
	BYTE s = 0;
	BYTE l = 0;
	
	BYTE hueh = 0;
	BYTE hues = 0;
	BYTE huel = 0;

	bool needHSLConversion = false;
	bool needMixColor = false;

	if ( _luminanceIntensity != 0 || _saturationIntensity != 100 )
		needHSLConversion = true;

	if ( _hueIntensity != 0 )
	{
		r = GetRValue(_hueColor);
		g = GetGValue(_hueColor);
		b = GetBValue(_hueColor);

		if ( r == b && b == g )
		{
			_hueIntensity = 0;
		}
		else
		{
			RGB2HSL(r, g, b, hueh, hues, huel);
			needHSLConversion = true;
		}
	}

	if ( _mixIntensity != 0 )
	{
		_mixIntensity = (int) min(255, _mixIntensity*2.55);
		needMixColor = true;
	}

	for(iy = 0; iy < height; iy++)
	{
		for(ix = 0; ix < width; ix++)
		{
			cl = GetPixel(destHDC, ix, iy);
			r = GetRValue(cl);
			g = GetGValue(cl);
			b = GetBValue(cl);

			if( !( r == 255 && g == 0 && b == 255 ) )
			{
				if ( needHSLConversion )
				{
					RGB2HSL(r, g, b, h, s, l);

					if(_hueIntensity != 0)
						h = (BYTE) max(0, min(240, (h*(100-_hueIntensity)+hueh*_hueIntensity)/100));
		
					if(_luminanceIntensity != 0)
						l = (BYTE) max(0, min(240, ((l*100)+(120*_luminanceIntensity))/100));

					if(_saturationIntensity != 100)
					{
						if ( _saturationIntensity > 100 )
							s = (BYTE) max(0, min(240, (240*(_saturationIntensity-100))/100));				
						else
							s = (BYTE) max(0, min(240, (s*_saturationIntensity)/100));
					}
				
					HSL2RGB(r, g, b, h, s, l);
				}

				if ( needMixColor )
				{
					// HueMix effect
					r = (BYTE)((r * (255 - _mixIntensity) + GetRValue(_mixColor) * _mixIntensity + 255) >> 8);
					g = (BYTE)((g * (255 - _mixIntensity) + GetGValue(_mixColor) * _mixIntensity + 255) >> 8);
					b = (BYTE)((b * (255 - _mixIntensity) + GetBValue(_mixColor) * _mixIntensity + 255) >> 8);
				}
			}
			else
			{
				transparencyNeeded = true;
				if ( !needHSLConversion && !needMixColor )
					return;
			}

			if ( needHSLConversion || needMixColor )
				SetPixelV(destHDC, ix, iy, RGB(r, g, b));
		}
	}
	//End Saturation / Hueing
}

