/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2006 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "StdAfx.h"
#include "SolidTexture.h"
#include "math.h"

SolidTexture::SolidTexture() : 
	m_Data(NULL),
	transparencyNeeded(false),
	useHover(false),
	usePressed(false)
{
}

SolidTexture::~SolidTexture()
{
	if ( m_Data != NULL )
		delete m_Data;

	Gradient.clear();
	HoverGradient.clear();
	PressedGradient.clear();
}

void SolidTexture::configure(const string &uniqueKey, const string &groupKey, bool alphaMap)
{
	NameValuePair gradientmodeConstants[] =
	{
		{ ".none", GRADIENT_NONE },
		{ ".horizontal", GRADIENT_HORIZONTAL },
		{ ".vertical", GRADIENT_VERTICAL },
		{ ".radial", GRADIENT_RADIAL },
		{ ".diagonal", GRADIENT_DIAGONAL },
		{ ".fdiagonal", GRADIENT_FDIAGONAL },
		{ ".bdiagonal", GRADIENT_BDIAGONAL },
		{ 0, 0 }
	};

	NameValuePair gradienttransformationConstants[] =
	{
		{ ".none", TRANSFORMATION_NONE },
		{ ".caricature", TRANSFORMATION_CHARICATURE },
		{ ".fisheye", TRANSFORMATION_FISHEYE },
		{ ".swirled", TRANSFORMATION_SWIRLED },
		{ ".cylinder", TRANSFORMATION_CYLINDER },
		{ ".shift", TRANSFORMATION_SHIFT },
		{ 0, 0 }
	};

	NameValuePair borderValues[] = 
	{
		{ ".none", 0 },
		{ "none", 0 },
		{ "raised", 1 },
		{ "sunken", 2 },
		{ "etched", 3 },
		{ "bump", 4 },
		{ 0, 0 }
	};

	string solidColorsTMP = "";

	string gradientColors = "";
	string hovergradientColors = "";

	string bevelCFG = "";
	string hoverbevelCFG = "";

	//Always Needed Default Values
	//Normal (for Hover) and Hover (for Pressed)
	//--------------------------------------------
	solidColorsTMP = Getx2RCLine(uniqueKey, groupKey, "SolidColors", "c0c0c0");

	parseCFG(solidColorsTMP, backgroundColor, lightBevelColor, darkBevelColor);
	//BG Insert in Gradient List
	Gradient.insert(Gradient.end(), backgroundColor );

	if (backgroundColor == RGB(255,0,255) || lightBevelColor == RGB(255,0,255) || darkBevelColor == RGB(255,0,255))
		transparencyNeeded = true;

	bevelCFG = hoverbevelCFG = Getx2RCLine(uniqueKey, groupKey, "SolidBevelSize", "0");
	parseBevelCFG(bevelCFG, bevelSizeLeft, bevelSizeRight, bevelSizeTop, bevelSizeBottom);

	gradientColors = hovergradientColors = Getx2RCLine(uniqueKey, groupKey, "SolidGradientColors", "");
	parseGradientCFG( gradientColors, 0 );

	verticalGradient = hoververticalGradient = Getx2RCBool(uniqueKey, groupKey, "SolidGradientVertical");
	repeatedGradient = hoverrepeatedGradient = Getx2RCInt(uniqueKey, groupKey, "SolidGradientRepeated", 0, 0);

	//Win2K/XP Special Gradient Modes
	gradientType = hovergradientType = Getx2RCNamedValue(uniqueKey, groupKey, "SolidGradientType", gradientmodeConstants, -1);
	transformationType = hovertransformationType = Getx2RCNamedValue(uniqueKey, groupKey, "SolidTransformationType", gradienttransformationConstants, -1);
	//--------------------------------------------

	borderMethod = hoverborderMethod = Getx2RCNamedValue(uniqueKey, groupKey, "BorderMethod", borderValues, 0);

	solidColorsTMP = Getx2RCLine(uniqueKey, groupKey, "HoverSolidColors", "");
	if (!solidColorsTMP.empty())
	{
		useHover = true;

		parseCFG(solidColorsTMP, hoverbackgroundColor, hoverlightBevelColor, hoverdarkBevelColor);
		//BG Insert in Gradient List
		HoverGradient.insert(HoverGradient.end(), hoverbackgroundColor );

		if (hoverbackgroundColor == RGB(255,0,255) || hoverlightBevelColor == RGB(255,0,255) || hoverdarkBevelColor == RGB(255,0,255))
			transparencyNeeded = true;

		hoverbevelCFG = Getx2RCLine(uniqueKey, groupKey, "HoverSolidBevelSize", bevelCFG);
		parseBevelCFG(hoverbevelCFG, hoverbevelSizeLeft, hoverbevelSizeRight, hoverbevelSizeTop, hoverbevelSizeBottom);

		hovergradientColors = Getx2RCLine(uniqueKey, groupKey, "HoverSolidGradientColors", gradientColors);
		parseGradientCFG( hovergradientColors, 1 );

		hoververticalGradient = Getx2RCBool(uniqueKey, groupKey, "HoverSolidGradientVertical", verticalGradient);
		hoverrepeatedGradient = Getx2RCInt(uniqueKey, groupKey, "HoverSolidGradientRepeated", repeatedGradient, 0);

		//Win2K/XP Special Gradient Modes
		hovergradientType = Getx2RCNamedValue(uniqueKey, groupKey, "HoverSolidGradientType", gradientmodeConstants, gradientType);
		hovertransformationType = Getx2RCNamedValue(uniqueKey, groupKey, "HoverSolidTransformationType", gradienttransformationConstants, transformationType);
	
		hoverborderMethod = Getx2RCNamedValue(uniqueKey, groupKey, "HoverBorderMethod", borderValues, borderMethod);
	}

	solidColorsTMP = Getx2RCLine(uniqueKey, groupKey, "PressedSolidColors", "");
	if (!solidColorsTMP.empty())
	{
		usePressed = true;

		parseCFG(solidColorsTMP, pressedbackgroundColor, pressedlightBevelColor, presseddarkBevelColor);
		//BG Insert in Gradient List
		PressedGradient.insert(PressedGradient.end(), pressedbackgroundColor );

		if (pressedbackgroundColor == RGB(255,0,255) || pressedlightBevelColor == RGB(255,0,255) || presseddarkBevelColor == RGB(255,0,255))
			transparencyNeeded = true;

		parseBevelCFG(Getx2RCLine(uniqueKey, groupKey, "PressedSolidBevelSize", hoverbevelCFG), pressedbevelSizeLeft, pressedbevelSizeRight, pressedbevelSizeTop, pressedbevelSizeBottom);

		parseGradientCFG(Getx2RCLine(uniqueKey, groupKey, "PressedSolidGradientColors", hovergradientColors), 2 );

		pressedverticalGradient = Getx2RCBool(uniqueKey, groupKey, "PressedSolidGradientVertical", hoververticalGradient);
		pressedrepeatedGradient = Getx2RCInt(uniqueKey, groupKey, "PressedSolidGradientRepeated", hoverrepeatedGradient, 0);

		//Win2K/XP Special Gradient Modes
		pressedgradientType = Getx2RCNamedValue(uniqueKey, groupKey, "PressedSolidGradientType", gradientmodeConstants, hovergradientType);
		pressedtransformationType = Getx2RCNamedValue(uniqueKey, groupKey, "PressedSolidTransformationType", gradienttransformationConstants, hovertransformationType);
	
		pressedborderMethod = Getx2RCNamedValue(uniqueKey, groupKey, "PressedBorderMethod", borderValues, hoverborderMethod);
	}
}

void SolidTexture::parseCFG(string cfg, COLORREF &bgcolor, COLORREF &lightcolor, COLORREF &darkcolor)
{
	while ( strstr( cfg.c_str(), "#" ) != NULL )
		cfg.replace(cfg.find("#"), 1, "", 0);
	
	int color;
	char main[32], light[32], dark[32];
	char *buffers[] = {main, light, dark};

	int numTokens = LCTokenize(cfg.c_str(), buffers, 3, NULL);
		
	if(numTokens == 1)
	{
		color = strtol(main, NULL, 16);
		bgcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
		lightcolor = GetSysColor(COLOR_3DLIGHT);
		darkcolor = GetSysColor(COLOR_3DSHADOW);
	}
	else if (numTokens == 2)
	{
		color = strtol(main, NULL, 16);
		bgcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

		color = strtol(light, NULL, 16);
		lightcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
		darkcolor = lightcolor;
	}
	else if (numTokens == 3)
	{
		color = strtol(main, NULL, 16);
		bgcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

		color = strtol(light, NULL, 16);
		lightcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

		color = strtol(dark, NULL, 16);
		darkcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
	}
}

void SolidTexture::parseGradientCFG(string cfg, int imageType)
{
	while ( strstr( cfg.c_str(), "#" ) != NULL )
		cfg.replace(cfg.find("#"), 1, "", 0);

	int color;
	COLORREF insertColor;
	char token[MAX_LINE_LENGTH];
	LPCSTR nextToken = cfg.c_str();
	
	int counter = 1;
	while (GetToken(nextToken, token, &nextToken, false))
	{
		color = strtol(token, NULL, 16);
		insertColor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
		if (imageType == 0)
		{
			Gradient.insert(Gradient.end(), insertColor );
		}
		else if (imageType == 1)
		{
			HoverGradient.insert(HoverGradient.end(), insertColor );
		}
		else if (imageType == 2)
		{
			PressedGradient.insert(PressedGradient.end(), insertColor );
		}
		counter++;
	}
}

void SolidTexture::parseBevelCFG(string cfg, int &leftBevel, int &rightBevel, int &topBevel, int &bottomBevel )
{
	char left[32], right[32], top[32], bottom[32];
	char *buffers[] = { left, right, top, bottom };

	leftBevel = 0;
	rightBevel = 0;
	topBevel = 0;
	bottomBevel = 0;

	int numTokens = LCTokenize(cfg.c_str(), buffers, 4, NULL);
		
	if(numTokens == 1)
	{
		leftBevel = atoi(left);
		rightBevel = leftBevel;
		topBevel = leftBevel;
		bottomBevel = leftBevel;
	}
	else if (numTokens == 2)
	{
		leftBevel = atoi(left);
		rightBevel = leftBevel;
		topBevel = atoi(right);
		bottomBevel = topBevel;
	}
	else if (numTokens == 4)
	{
		leftBevel = atoi(left);
		rightBevel = atoi(right);
		topBevel = atoi(top);
		bottomBevel = atoi(bottom);
	}
}

void SolidTexture::apply(HDC hDC, int x, int y, int width, int height, int transparencyMode, int imageType)
{
	if ( width <= 0 || height <= 0 )
		return;

	RECT r;
	r.left = x;
	r.top = y;
	r.right = r.left + width;
	r.bottom = r.top + height;

	int bevelWidth = 0;
	int bevelHeight = 0;

	if ( imageType == 2 && !usePressed )
		imageType = 1;
	if ( imageType == 1 && !useHover )
		imageType = 0;

	if (imageType == 0)
	{
		bevelWidth = bevelSizeLeft+bevelSizeRight;
		bevelHeight = bevelSizeTop+bevelSizeBottom;
		
		// fill in the background
		if (bevelWidth+bevelHeight > 0 && (bevelWidth < width && bevelHeight < height))
		{
			::SetBkColor(hDC, darkBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.right = r.right-bevelSizeRight;
			r.bottom = r.bottom-bevelSizeBottom;
			::SetBkColor(hDC, lightBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.left = r.left+bevelSizeLeft;
			r.top = r.top+bevelSizeTop;

			if ( Gradient.size() > 1 )
			{
				if ( gradientType == -1 || !IsOS(OS_2KXP) )
					solidGradientPaint(hDC, Gradient, r, imageType);
				else
					DrawGradient(hDC, r, Gradient[0], Gradient[1], gradientType, transformationType);
			}
			else
			{
				::SetBkColor(hDC, backgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
		else
		{
			if ( Gradient.size() > 1 )
			{
				if ( gradientType == -1 || !IsOS(OS_2KXP) )
					solidGradientPaint(hDC, Gradient, r, imageType);
				else
					DrawGradient(hDC, r, Gradient[0], Gradient[1], gradientType, transformationType);
			}
			else
			{
				::SetBkColor(hDC, backgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
	}
	else if (imageType == 1)
	{
		bevelWidth = hoverbevelSizeLeft+hoverbevelSizeRight;
		bevelHeight = hoverbevelSizeTop+hoverbevelSizeBottom;

		// fill in the background
		if (bevelWidth+bevelHeight > 0 && (bevelWidth < width && bevelHeight < height))
		{
			::SetBkColor(hDC, hoverdarkBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.right = r.right-hoverbevelSizeRight;
			r.bottom = r.bottom-hoverbevelSizeBottom;
			::SetBkColor(hDC, hoverlightBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.left = r.left+hoverbevelSizeLeft;
			r.top = r.top+hoverbevelSizeTop;

			if ( HoverGradient.size() > 1 )
			{
				if ( hovergradientType == -1 || !IsOS(OS_2KXP) )
					solidGradientPaint(hDC, HoverGradient, r, imageType);
				else
					DrawGradient(hDC, r, HoverGradient[0], HoverGradient[1], hovergradientType, hovertransformationType);
			}
			else
			{
				::SetBkColor(hDC, hoverbackgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
		else
		{
			if ( HoverGradient.size() > 1 )
			{
				if ( hovergradientType == -1 || !IsOS(OS_2KXP) )
					solidGradientPaint(hDC, HoverGradient, r, imageType);
				else
					DrawGradient(hDC, r, HoverGradient[0], HoverGradient[1], hovergradientType, hovertransformationType);
			}
			else
			{
				::SetBkColor(hDC, hoverbackgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
	}
	else if (imageType == 2)
	{
		bevelWidth = pressedbevelSizeLeft+pressedbevelSizeRight;
		bevelHeight = pressedbevelSizeTop+pressedbevelSizeBottom;

		// fill in the background
		if (bevelWidth+bevelHeight > 0 && (bevelWidth < width && bevelHeight < height))
		{
			::SetBkColor(hDC, presseddarkBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.right = r.right-pressedbevelSizeRight;
			r.bottom = r.bottom-pressedbevelSizeBottom;
			::SetBkColor(hDC, pressedlightBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.left = r.left+pressedbevelSizeLeft;
			r.top = r.top+pressedbevelSizeTop;

			if ( PressedGradient.size() > 1 )
			{
				if ( pressedgradientType == -1 || !IsOS(OS_2KXP) )
					solidGradientPaint(hDC, PressedGradient, r, imageType);
				else
					DrawGradient(hDC, r, PressedGradient[0], PressedGradient[1], pressedgradientType, pressedtransformationType);
			}
			else
			{
				::SetBkColor(hDC, pressedbackgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
		else
		{
			if ( PressedGradient.size() > 1 )
			{
				if ( pressedgradientType == -1 || !IsOS(OS_2KXP) )
					solidGradientPaint(hDC, PressedGradient, r, imageType);
				else
					DrawGradient(hDC, r, PressedGradient[0], PressedGradient[1], pressedgradientType, pressedtransformationType);
			}
			else
			{
				::SetBkColor(hDC, pressedbackgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
	}

	RECT completeFillRect = {x, y, x+width, y+height};

	if ( (borderMethod == 1 && imageType == 0) || (hoverborderMethod == 1 && imageType == 1) || (pressedborderMethod == 1 && imageType == 2) )
		DrawEdge(hDC, &completeFillRect, EDGE_RAISED, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 2 && imageType == 0) || (hoverborderMethod == 2 && imageType == 1) || (pressedborderMethod == 2 && imageType == 2) )
		DrawEdge(hDC, &completeFillRect, EDGE_SUNKEN, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 3 && imageType == 0) || (hoverborderMethod == 3 && imageType == 1) || (pressedborderMethod == 3 && imageType == 2) )
		DrawEdge(hDC, &completeFillRect, EDGE_ETCHED, BF_RECT | BF_SOFT);
	else if ( (borderMethod == 4 && imageType == 0) || (hoverborderMethod == 4 && imageType == 1) || (pressedborderMethod == 4 && imageType == 2) )
		DrawEdge(hDC, &completeFillRect, EDGE_BUMP, BF_RECT | BF_SOFT);
}

void SolidTexture::applyOverlay(HDC hDC, int x, int y, int width, int height, int alphaValue, int imageType)
{
	if ( width <= 0 || height <= 0 )
		return;

	HDC bufferDC = CreateCompatibleDC(NULL);

	BITMAPINFO alpha;
	ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
	alpha.bmiHeader.biWidth=width;      // Set size you need
	alpha.bmiHeader.biHeight=height;    // Set size you need
	alpha.bmiHeader.biPlanes=1;
	alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
	alpha.bmiHeader.biSizeImage=0;
	alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	alpha.bmiHeader.biClrUsed= 0;
	alpha.bmiHeader.biClrImportant= 0;
	VOID *pvBits;

	HBITMAP bufferBitmap = CreateDIBSection( bufferDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

	HBITMAP oldBMP = (HBITMAP) SelectObject(bufferDC, bufferBitmap);

	apply(bufferDC, 0, 0, width, height, 1, imageType);

	BITMAP bm;
	GetObject(bufferBitmap, sizeof(BITMAP), &bm);

	BYTE *pPixel = (LPBYTE)bm.bmBits;

	int iSize = bm.bmHeight * bm.bmWidth;

	for (int j = 0; j < iSize; ++j)
	{
		//Convert to Full Opacity
		pPixel[3] = 255;
		
		//Remove Magic Pink
		if ( pPixel[0] == 255 && pPixel[1] == 0 && pPixel[2] == 255 )
		{
			pPixel[0] = 0;
			pPixel[1] = 0;
			pPixel[2] = 0;
			pPixel[3] = 0;
		}

		pPixel+= 4;
	}

	BLENDFUNCTION blendFunction = { AC_SRC_OVER, 0, alphaValue, AC_SRC_ALPHA };

	AlphaBlend(hDC, x, y, width, height, bufferDC, 0, 0, width, height, blendFunction);
		
	bufferBitmap = (HBITMAP) SelectObject(bufferDC, oldBMP);
	DeleteObject(bufferBitmap);
	DeleteDC(bufferDC);
}

void SolidTexture::solidGradientPaint(HDC destHDC, vector<COLORREF> workGradient, RECT r, int imageType)
{
	bool vertical = false;
	int repeated = 0;
	if ( imageType == 0 )
	{
		vertical = verticalGradient;
		repeated = repeatedGradient;
	}
	else if( imageType == 1 )
	{
		vertical = hoververticalGradient;
		repeated = hoverrepeatedGradient;
	}
	else if( imageType == 2 )
	{
		vertical = pressedverticalGradient;
		repeated = pressedrepeatedGradient;
	}
	
	if ( repeated == 0 )
	{
		int nWidth = r.right-r.left;
		int nHeight = r.bottom-r.top;

		RECT r2 = {0,0,0,0};
		double cr, cg, cb, dr, dg, db;

		// ColorsSteps between Start and End Color
		int gradSteps = 0;
		int unequalRest = 0;

		int colorCount = workGradient.size();

		if (vertical)
		{
			gradSteps = max(0, (nHeight-colorCount)/(colorCount-1));
			unequalRest = (nHeight-colorCount)%(colorCount-1);
		}
		else
		{
			gradSteps = max(0, (nWidth-colorCount)/(colorCount-1));
			unequalRest = (nWidth-colorCount)%(colorCount-1);
		}

		int lastGradientPos = 0;
		if (vertical)
			lastGradientPos = r.top;
		else
			lastGradientPos = r.left;

		int counter = 0;
		for (ColorRefIterator cit = workGradient.begin(); cit != workGradient.end(); cit++ )
		{
			if (cit == workGradient.end())
				break;

			counter++;

			COLORREF one = *cit;
			cit++;
			COLORREF two = *cit;
			cit--;
		
			// get current color
			cr = GetRValue( one );
			cg = GetGValue( one );
			cb = GetBValue( one );

			// Get delta color (+1 cause we must reach the Final Color)
			dr = (cr - GetRValue( two )) / (gradSteps+1);
			dg = (cg - GetGValue( two )) / (gradSteps+1);
			db = (cb - GetBValue( two )) / (gradSteps+1);

			//Paint Real Gradient Color
			//-------------------------------------------
			if ( unequalRest != 0 && counter > 1 )
			{
				unequalRest--;
				if (vertical)
				{
					SetRect(&r2, r.left, lastGradientPos, r.right, min(lastGradientPos+1, r.bottom));
				}
				else
				{
					SetRect(&r2, lastGradientPos, r.top, min(lastGradientPos+1, r.right), r.bottom);
				}

				if (r2.right-r2.left > 0 && r2.bottom-r2.top > 0)
				{
					::SetBkColor(destHDC, one);
					::ExtTextOut(destHDC, 0, 0, ETO_OPAQUE, &r2, NULL, 0, NULL);
				}
				
				if (vertical)
					lastGradientPos = r2.bottom;
				else
					lastGradientPos = r2.right;
			}
			//-------------------------------------------

			for (int i = 0; i <= gradSteps; i++)
			{
				if (vertical)
				{
					SetRect(&r2, r.left, lastGradientPos+i, r.right, min(lastGradientPos+i+1, r.bottom));
				}
				else
				{
					SetRect(&r2, lastGradientPos+i, r.top, min(lastGradientPos+i+1, r.right), r.bottom);
				}

				if (r2.right-r2.left > 0 && r2.bottom-r2.top > 0)
				{
					::SetBkColor(destHDC, RGB((int)cr, (int)cg, (int)cb));
					::ExtTextOut(destHDC, 0, 0, ETO_OPAQUE, &r2, NULL, 0, NULL);
				}
				
				cr -= dr;
				cg -= dg;
				cb -= db;
			}

			if (vertical)
				lastGradientPos = r2.bottom;
			else
				lastGradientPos = r2.right;
		}
	}
	else
	{
		int nWidth = r.right-r.left;
		int nHeight = r.bottom-r.top;

		RECT r2 = {0,0,0,0};
		COLORREF color;

		int lastGradientPos = 0;
		if (vertical)
			lastGradientPos = r.top;
		else
			lastGradientPos = r.left;

		int counter = 0;
		int nSize = 0;
		
		if (vertical)
			nSize = nHeight;
		else
			nSize = nWidth; 

		while ( counter < nSize )
		{
			for (ColorRefIterator cit = workGradient.begin(); cit != workGradient.end(); cit++ )
			{
				color = *cit;

				if (vertical)
				{
					SetRect(&r2, r.left, lastGradientPos, r.right, min(lastGradientPos+repeated, r.bottom));
				}
				else
				{
					SetRect(&r2, lastGradientPos, r.top, min(lastGradientPos+repeated, r.right), r.bottom);
				}

				if (r2.right-r2.left > 0 && r2.bottom-r2.top > 0 && color != RGB(255,0,255) )
				{
					::SetBkColor(destHDC, color);
					::ExtTextOut(destHDC, 0, 0, ETO_OPAQUE, &r2, NULL, 0, NULL);
				}

				if (vertical)
					lastGradientPos = r2.bottom;
				else
					lastGradientPos = r2.right;

				counter++;

				if ( counter == nSize )
					break;
			}
		}
	}
}

void SolidTexture::DrawGradient( HDC hDC, RECT& rect, COLORREF startColor, COLORREF endColor, int gradientType, int transformationType)
{
	int i, j, y, x1, y1;
	double r, d, n1;

	// Calculate image size
	int width = rect.right - rect.left;
	int height = rect.bottom - rect.top;
	int size = height*width*4;
	double radius = sqrt( pow(height/2,2) + pow(width/2,2) );
	double k = double(rect.bottom-rect.top)/double(rect.right-rect.left);
	if ( gradientType == GRADIENT_BDIAGONAL )
		k = double(rect.bottom-rect.top)/double(rect.left-rect.right);
	if ( ( gradientType == GRADIENT_FDIAGONAL ) || ( gradientType == GRADIENT_BDIAGONAL ) )
		gradientType = GRADIENT_DIAGONAL;
	double n = 0.0;
	double k1 = -1/k;

	// Create data buffer
	m_Data = new unsigned char[size];
	
	// Create gradient
	for ( i=0; i<height; i++ )
	{
		for ( j=0; j<width; j++ )
		{
			switch (gradientType)
			{
				// Horizontal gradient
				case GRADIENT_HORIZONTAL:
				{
					m_Data[i*width*4 + j*4] = (unsigned char)(((double)j/(double)width)*GetBValue(endColor) + (1-(double)j/(double)width)*GetBValue(startColor));
					m_Data[i*width*4 + j*4+1] = (unsigned char)(((double)j/(double)width)*GetGValue(endColor) + (1-(double)j/(double)width)*GetGValue(startColor));
					m_Data[i*width*4 + j*4+2] = (unsigned char)(((double)j/(double)width)*GetRValue(endColor) + (1-(double)j/(double)width)*GetRValue(startColor));
				}
				break;

				// Vertical gradient
				case GRADIENT_VERTICAL:
				{
					m_Data[i*width*4 + j*4] = (unsigned char)(((double)i/(double)height)*GetBValue(endColor) + (1-(double)i/(double)height)*GetBValue(startColor));
					m_Data[i*width*4 + j*4+1] = (unsigned char)(((double)i/(double)height)*GetGValue(endColor) + (1-(double)i/(double)height)*GetGValue(startColor));
					m_Data[i*width*4 + j*4+2] = (unsigned char)(((double)i/(double)height)*GetRValue(endColor) + (1-(double)i/(double)height)*GetRValue(startColor));
				}
				break;

				// Radial gradient
				case GRADIENT_RADIAL:
				{
					r = sqrt( pow(double(i-height/2)/radius,2) + pow(double(j-width/2)/radius,2) );
					if ( r <= 1.0 )
					{
						m_Data[i*width*4 + j*4] = (unsigned char)(r*GetBValue(endColor) + (1-r)*GetBValue(startColor));
						m_Data[i*width*4 + j*4+1] = (unsigned char)(r*GetGValue(endColor) + (1-r)*GetGValue(startColor));
						m_Data[i*width*4 + j*4+2] = (unsigned char)(r*GetRValue(endColor) + (1-r)*GetRValue(startColor));
					}
					else
					{
						m_Data[i*width*4 + j*4] = GetBValue(endColor);
						m_Data[i*width*4 + j*4+1] = GetGValue(endColor);
						m_Data[i*width*4 + j*4+2] = GetRValue(endColor);
					}
				}
				break;

				// Diagonal gradient
				case GRADIENT_DIAGONAL:
				{
					y = int(k*(j-width/2) + n);
					n1 = int((i-height/2) - k1*(j-width/2));
					x1 = int((n1-n)/(k-k1));
					y1 = int(k1*x1 + n1);
					d = sqrt( pow((j-width/2)-x1,2) + pow((i-height/2)-y1,2) ) / (2*radius);
					if ( (i-height/2) > y )
					{
						m_Data[i*width*4 + j*4] = (unsigned char)((0.5-d)*GetBValue(endColor) + (0.5+d)*GetBValue(startColor));
						m_Data[i*width*4 + j*4+1] = (unsigned char)((0.5-d)*GetGValue(endColor) + (0.5+d)*GetGValue(startColor));
						m_Data[i*width*4 + j*4+2] = (unsigned char)((0.5-d)*GetRValue(endColor) + (0.5+d)*GetRValue(startColor));
					}
					else
					{
						m_Data[i*width*4 + j*4] = (unsigned char)((0.5-d)*GetBValue(startColor) + (0.5+d)*GetBValue(endColor));
						m_Data[i*width*4 + j*4+1] = (unsigned char)((0.5-d)*GetGValue(startColor) + (0.5+d)*GetGValue(endColor));
						m_Data[i*width*4 + j*4+2] = (unsigned char)((0.5-d)*GetRValue(startColor) + (0.5+d)*GetRValue(endColor));
					}
				}
				break;
			}
		}
	}

	if ( transformationType != TRANSFORMATION_NONE )
		ApplyTransformation( width, height, transformationType );

	// Create temporary bitmap
	HBITMAP hBitmap = (HBITMAP)CreateBitmap( width, height, 1, 32, m_Data );

	// Create temporary device context
	HDC hMemDC = CreateCompatibleDC(hDC);
	HBITMAP hOldBitmap = (HBITMAP)SelectObject( hMemDC, hBitmap );

	// Blit gradient bitmap
	BitBlt( hDC, rect.left, rect.top, width, height, hMemDC, 0, 0, SRCCOPY );

	// Free memory
	SelectObject( hMemDC, hOldBitmap );
	DeleteDC(hMemDC);
	DeleteObject(hBitmap);
	delete m_Data;
	m_Data = NULL;
}

void SolidTexture::ApplyTransformation(int width, int height, int transformationType)
{
	int X, Y;
	double r, fi;
	double R = sqrt( pow(width/4,2) + pow(height/4,2) );
	double FI = atan( double(height)/double(width) );

	int size = height*width*4;
	unsigned char* newBuffer = new unsigned char[size];
	memcpy( newBuffer, m_Data, size );

	for ( int i=0; i<height; i++ )
	{
		for ( int j=0; j<width; j++ )
		{
			switch (transformationType)
			{
				case TRANSFORMATION_CHARICATURE:
				{
					fi = atan(double(i-height/2)/double(j-width/2));
					r = sqrt( pow((i-height/2),2) + pow((j-width/2),2) );
					X = int(sqrt(r*R*2)*cos(fi));
					Y = int(fabs(sqrt(r*R*2)*sin(fi)));
					newBuffer[i*width*4 + j*4] = m_Data[Y*width*4 + X*4];
					newBuffer[i*width*4 + j*4+1] = m_Data[Y*width*4 + X*4+1];
					newBuffer[i*width*4 + j*4+2] = m_Data[Y*width*4 + X*4+2];
				}
				break;

				case TRANSFORMATION_FISHEYE:
				{
					fi = atan(double(i-height/2)/double(j-width/2));
					r = sqrt( pow((i-height/2),2) + pow((j-width/2),2) );
					X = int((pow(r,2)/R)*cos(fi));
					Y = int(fabs((pow(r,2)/R)*sin(fi))) % height;
					newBuffer[i*width*4 + j*4] = m_Data[Y*width*4 + X*4];
					newBuffer[i*width*4 + j*4+1] = m_Data[Y*width*4 + X*4+1];
					newBuffer[i*width*4 + j*4+2] = m_Data[Y*width*4 + X*4+2];
				}
				break;

				case TRANSFORMATION_SWIRLED:
				{
					fi = atan(double(i-height/2)/double(j-width/2));
					r = sqrt( pow((i-height/2),2) + pow((j-width/2),2) );
					X = int(fabs(r*cos(fi+r/8)));
					Y = int(fabs(r*sin(fi+r/8)));
					newBuffer[i*width*4 + j*4] = m_Data[Y*width*4 + X*4];
					newBuffer[i*width*4 + j*4+1] = m_Data[Y*width*4 + X*4+1];
					newBuffer[i*width*4 + j*4+2] = m_Data[Y*width*4 + X*4+2];
				}
				break;

				case TRANSFORMATION_CYLINDER:
				{
					fi = atan(double(i-height/2)/double(j-width/2));
					r = sqrt( pow((i-height/2),2) + pow((j-width/2),2) );
					X = abs((j-width/2) + int(fi+r/10)%32 - 8);
					Y = i;
					newBuffer[i*width*4 + j*4] = m_Data[Y*width*4 + X*4];
					newBuffer[i*width*4 + j*4+1] = m_Data[Y*width*4 + X*4+1];
					newBuffer[i*width*4 + j*4+2] = m_Data[Y*width*4 + X*4+2];
				}
				break;

				case TRANSFORMATION_SHIFT:
				{
					fi = atan(double(i-height/2)/double(j-width/2));
					r = sqrt( pow((i-height/2),2) + pow((j-width/2),2) );
					X = int(fabs(r*cos(fi)));
					Y = int(fabs(r*sin(fi)));
					newBuffer[i*width*4 + j*4] = m_Data[Y*width*4 + X*4];
					newBuffer[i*width*4 + j*4+1] = m_Data[Y*width*4 + X*4+1];
					newBuffer[i*width*4 + j*4+2] = m_Data[Y*width*4 + X*4+2];
				}
				break;
			}
		}
	}

	memcpy( m_Data, newBuffer, size );
	delete newBuffer;
}