/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2006 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#if !defined(__SOLIDTEXTURE_H)
#define __SOLIDTEXTURE_H

#include "Texture.h"

typedef vector<COLORREF>::iterator ColorRefIterator;

#define	GRADIENT_NONE			-1
#define	GRADIENT_HORIZONTAL		0
#define	GRADIENT_VERTICAL		1
#define	GRADIENT_RADIAL			2
#define	GRADIENT_DIAGONAL		3
#define	GRADIENT_FDIAGONAL		4
#define	GRADIENT_BDIAGONAL		5

#define	TRANSFORMATION_NONE			-1
#define	TRANSFORMATION_CHARICATURE  0
#define	TRANSFORMATION_FISHEYE		1
#define	TRANSFORMATION_SWIRLED		2
#define	TRANSFORMATION_CYLINDER		3
#define	TRANSFORMATION_SHIFT		4

class SolidTexture : public Texture
{
public:

	SolidTexture();
	virtual ~SolidTexture();

	void configure(const string &uniqueKey, const string &groupKey, bool alphaMap);
	void apply(HDC hDC, int x, int y, int width, int height, int transparencyMode, int imageType);
	void applyOverlay(HDC hDC, int x, int y, int width, int height, int alphaValue, int imageType);
	void applyAnim(HDC hDC, int x, int y, int width, int height, int transparencyMode, int imageType, int count) { return; };

	bool isTransparencyNeeded() { return transparencyNeeded; };
	int isTextureType() { return 3; };
	bool useHoverState() { return useHover; };
	bool usePressedState() { return usePressed; };

private:

	bool transparencyNeeded;

	bool useHover;
	bool usePressed;

	void parseCFG(string cfg, COLORREF &bgcolor, COLORREF &lightcolor, COLORREF &darkcolor);
	void parseGradientCFG(string cfg, int imageType);
	void parseBevelCFG(string cfg, int &leftBevel, int &rightBevel, int &topBevel, int &bottomBevel);

	void solidGradientPaint(HDC destHDC, vector<COLORREF> workGradient, RECT r, int imageType);

	unsigned char* m_Data;
	void DrawGradient( HDC hDC, RECT& rect, COLORREF startColor, COLORREF endColor, int gradientType, int transformationType);
	void ApplyTransformation(int width, int height, int transformationType);

	int bevelSizeLeft, bevelSizeRight, bevelSizeTop, bevelSizeBottom;
	int hoverbevelSizeLeft, hoverbevelSizeRight, hoverbevelSizeTop, hoverbevelSizeBottom;
	int pressedbevelSizeLeft, pressedbevelSizeRight, pressedbevelSizeTop, pressedbevelSizeBottom;

	bool verticalGradient;
	bool hoververticalGradient;
	bool pressedverticalGradient;
	
	int repeatedGradient;
	int hoverrepeatedGradient;
	int pressedrepeatedGradient;

	int gradientType;
	int hovergradientType;
	int pressedgradientType;

	int transformationType;
	int hovertransformationType;
	int pressedtransformationType;

	COLORREF backgroundColor;
	COLORREF darkBevelColor;
	COLORREF lightBevelColor;

	COLORREF hoverbackgroundColor;
	COLORREF hoverdarkBevelColor;
	COLORREF hoverlightBevelColor;

	COLORREF pressedbackgroundColor;
	COLORREF presseddarkBevelColor;
	COLORREF pressedlightBevelColor;

	vector<COLORREF> Gradient;
	vector<COLORREF> HoverGradient;
	vector<COLORREF> PressedGradient;

	int borderMethod;
	int hoverborderMethod;
	int pressedborderMethod;
};

#endif
