/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2006 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#if !defined(__UTILITY_H)
#define __UTILITY_H

class xFont;
class Texture;
class GhostLabel;

const int MAX_INTEGER = 0x7FFFFFFF;
const int MIN_INTEGER = 0x80000000;

#define OS_9X   0x0001
#define OS_NT   0x0002
#define OS_95   0x0004
#define OS_98   0x0008
#define OS_ME   0x0010
#define OS_NT4  0x0020
#define OS_2K   0x0040
#define OS_XP   0x0080
#define OS_2KXP 0x0100

#define DEFAULT_PREFIX "AllLabels"

enum
{
	MULTIBLT_STRETCH,
	MULTIBLT_TILE,
	MULTIBLT_TILEHORIZONTAL,
	MULTIBLT_TILEVERTICAL
};

struct NameValuePair
{
	const char *name;
	int value;
};

// Find windows at any level (not only top level windows)
struct FindWindowEx2Data 
{
	LPCSTR pszClass;
	LPCSTR pszTitle;
	HWND *phWnd;
};

void getMonitorInfo();

BOOL CALLBACK EnumMonitorsProc(HMONITOR hMonitor, HDC hdc, LPRECT lpRMonitor, LPARAM dwData);

HWND FindWindowEx2(string pszClass, string pszTitle);
static BOOL CALLBACK FindWindowEx2EnumWindowsProc(HWND hWnd, LPARAM lParam);

HINSTANCE xLSExecute(HWND Owner, LPCSTR szCommand, int nShowCmd);

HWND GetLitestepDesktop();

bool IsAppWindow(const HWND &hWnd);
bool IsOS(const int &os);
string QueryPixelDepth();
string trimit(const string &toTrim);
void PrefixedLSSetVariable(string prefix, string var, string value);

void ModifyStyle(HWND hWnd, DWORD removeStyle, DWORD addStyle);
void ModifyStyleEx(HWND hWnd, DWORD removeStyle, DWORD addStyle);

int SplitString(const string &process, string &one, string &two, string &three, string &four, string &five, string &six, string &seven);

bool GetxRCBool(const string &uniqueKey, const string &baseName, bool defaultVal = false);
bool Getx2RCBool(const string &uniqueKey, const string &groupKey, const string &baseName, bool defaultVal = false);

string GetxRCString(const string &uniqueKey, const string &baseName, const string &defaultVal);
string Getx2RCString(const string &uniqueKey, const string &groupKey, const string &baseName, const string &defaultVal);

string GetxRCLine(const string &uniqueKey, const string &baseName, const string &defaultVal);
string Getx2RCLine(const string &uniqueKey, const string &groupKey, const string &baseName, const string &defaultVal);

StringList GetxRCLineList(const string &uniqueKey);

int GetxRCNamedValue(const string &uniqueKey, const string &baseName, const NameValuePair *nameValuePairs, int defaultVal);
int Getx2RCNamedValue(const string &uniqueKey, const string &groupKey, const string &baseName, const NameValuePair *nameValuePairs, int defaultVal);

StringList GetxRCNameList(const string &uniqueKey);

int GetxRCColor(const string &uniqueKey, const string &baseName, int defaultVal);
int Getx2RCColor(const string &uniqueKey, const string &groupKey, const string &baseName, int defaultVal);

int GetxRCInt(const string &uniqueKey, const string &baseName, int defaultVal, int minVal = MIN_INTEGER, int maxVal = MAX_INTEGER);
int Getx2RCInt(const string &uniqueKey, const string &groupKey, const string &baseName, int defaultVal, int minVal = MIN_INTEGER, int maxVal = MAX_INTEGER);

int xParseCoord(const string &aString, int defaultVal, int maxVal, HWND box, bool vertical = false);
int GetxRCCoord(const string &uniqueKey, const string &baseName, int defaultVal, int maxVal, HWND box, bool vertical = false);
int Getx2RCCoord(const string &uniqueKey, const string &groupKey, const string &baseName, int defaultVal, int maxVal, HWND box, bool vertical = false);

int xParseDimen(const string &aString, int defaultVal, int maxVal, HWND box, bool vertical = false);
int GetxRCDimen(const string &uniqueKey, const string &baseName, int defaultVal, int maxVal, HWND box, bool vertical = false);
int Getx2RCDimen(const string &uniqueKey, const string &groupKey, const string &baseName, int defaultVal, int maxVal, HWND box, bool vertical = false);

xFont *GetRCFont(const string &uniqueKey, const string &groupKey);

Texture *GetRCTexture(const string &uniqueKey, const string &groupKey, bool externalConfigure = false);

void PaintDesktopEx(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, int xSrc, int ySrc, BOOL updateCache);

void TileBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, DWORD rasterOp);
void HorizontalTileBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, DWORD rasterOp);
void VerticalTileBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, DWORD rasterOp);
void MultiBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, int cxLeft, int cyTop, int cxRight, int cyBottom, int mode, DWORD rasterOp);

LPTSTR RegQueryStringValue(HKEY hKey, LPCTSTR pszSubKey, LPCTSTR pszValueName, LPTSTR pszBuffer, int nBufferSize, LPTSTR pszDefaultValue);

vector<string> split(string str, string sep);

BOOL GetClipboardText(HWND hwndOwner, LPTSTR pszBuffer, UINT cchBuffer);
BOOL SetClipboardText(HWND hwndOwner, LPCTSTR pszText);

string IntToString(int input);

#endif
