/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2006 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "StdAfx.h"
#include "Label.h"
#include "xFont.h"
#include "GhostLabel.h"
#include "SystemInfo.h"

#include "VerInfo.h"

#define TIMER_PLAYSOUND		111

//Alpha Transparency
//-----------------------------------------------------
#ifndef WS_EX_LAYERED
	#define WS_EX_LAYERED           0x00080000L
	#define LWA_COLORKEY            0x00000001L
	#define ULW_COLORKEY            0x00000001L
	#define LWA_ALPHA               0x00000002L
	#define ULW_ALPHA               0x00000002L
	#define ULW_OPAQUE              0x00000003L
#endif // WS_EX_LAYERED

//DropShadow under XP
#ifndef CS_DROPSHADOW
	#define CS_DROPSHADOW 0x00020000
#endif

//TextEditBox
#define LM_TEXTEDITBOX 9650

//PlaySound
#define   NOCOMPMAN     //  - dont include COMPMAN
#define   NODRAWDIB   //    - dont include DRAWDIB
#define   NOVIDEO     //    - dont include video capture interface
#define   NOAVIFMT    //    - dont include AVI file format structs
#define   NOMMREG     //    - dont include MMREG
#define   NOAVIFILE   //    - dont include AVIFile interface
//#define   NOMCIWND    //    - dont include AVIWnd class.
#define   NOAVICAP    //    - dont include AVICap class.
#define   NOMSACM     //    - dont include ACM stuff.
#include "vfw.h"

// Bang commands
void BangAlwaysOnTop(HWND hwndCaller, LPCSTR pszArgs);
void BangClipboardCopy(HWND hwndCaller, LPCSTR pszArgs);
void BangClipboardPaste(HWND hwndCaller, LPCSTR pszArgs);
void BangCreate(HWND hwndCaller, LPCSTR pszArgs);
void BangDestroy(HWND hwndCaller, LPCSTR pszArgs);
void BangDock(HWND hwndCaller, LPCSTR pszArgs);
void BangFocus(HWND hwndCaller, LPCSTR pszArgs);
void BangGhosted(HWND hwndCaller, LPCSTR pszArgs);
void BangHide(HWND hwndCaller, LPCSTR pszArgs);
void BangLSBoxHook(HWND hwndCaller, LPCSTR pszArgs);
void BangModuleHook(HWND hwndCaller, LPCSTR pszArgs);
void BangMove(HWND hwndCaller, LPCSTR pszArgs);
void BangMoveBy(HWND hwndCaller, LPCSTR pszArgs);
void BangNext(HWND hwndCaller, LPCSTR pszArgs);
void BangPrevious(HWND hwndCaller, LPCSTR pszArgs);
void BangRefresh(HWND hwndCaller, LPCSTR pszArgs);
void BangReposition(HWND hwndCaller, LPCSTR pszArgs);
void BangRepositionBy(HWND hwndCaller, LPCSTR pszArgs);
void BangResize(HWND hwndCaller, LPCSTR pszArgs);
void BangResizeBy(HWND hwndCaller, LPCSTR pszArgs);
void BangScroll(HWND hwndCaller, LPCSTR pszArgs);
void BangSetAlpha(HWND hwndCaller, LPCSTR pszArgs);
void BangSetAnimation(HWND hwndCaller, LPCSTR pszArgs);
void BangSetText(HWND hwndCaller, LPCSTR pszArgs);
void BangShow(HWND hwndCaller, LPCSTR pszArgs);
void BangShowHide(HWND hwndCaller, LPCSTR pszArgs);
void BangToggle(HWND hwndCaller, LPCSTR pszArgs);
void BangUpdate(HWND hwndCaller, LPCSTR pszArgs);

void BangInfoExport(HWND hwndCaller, LPCSTR pszArgs);

void BangParseEvars(HWND hwndCaller, LPCSTR pszArgs);
void BangSetEvars(HWND hwndCaller, LPCSTR pszArgs);

void BangPlaySound(HWND hwndCaller, LPCSTR pszArgs);

void BangTexteditBoxExecute(HWND hwndCaller, LPCSTR pszArgs);

HWND m_hWndSoundFile = NULL;

HINSTANCE hInstance;
HWND messageHandler;

string OSVersion = "";
string OSExtendedVersion = "";

int screenWidth = 0;
int screenHeight = 0;
int minscreenX = 0;
int	minscreenY = 0;
int maxscreenWidth = 0;
int	maxscreenHeight = 0;

int bitsperpixel = atoi( QueryPixelDepth().c_str() );

int realMonitors = 1;
int monitorIndex = 0;
MONITORINFO monitorInfoArray[3];
SIZE monitorSizeArray[3];

UINT (WINAPI *_PrivateExtractIcons)(LPCTSTR,int,int,int,HICON*,UINT*,UINT,UINT) = NULL;

LabelList labelList;

int lsMessages[] = {
	LM_GETREVID,
	LM_REFRESH,
	0
};

Label *lookupLabel( const string &name );

#define LM_UPDATEBG (WM_USER + 1)

LRESULT WINAPI MessageHandlerProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
		case LM_GETREVID:
		{
			UINT uLength;
			StringCchPrintf((char*)lParam, 64, "%s.dll: %s", V_NAME, V_VERSION);
			
			if (SUCCEEDED(StringCchLength((char*)lParam, 64, &uLength)))
				return uLength;

			lParam = NULL;
			return 0;
		}
		
		case LM_REFRESH:
		{
			StringList labelNames = GetxRCNameList("*label");

			for(LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if(!(*lit)->getBox())
				{
					// destroy all labels that no longer exist and that are not in a box
					for(StringListIterator sit = labelNames.begin(); sit != labelNames.end(); sit++)
					{
						if(_stricmp((*sit).c_str(), (*lit)->getName().c_str()) == NULL)
							break;
					}
					if (sit == labelNames.end())
					{
						(*lit)->hide();
						labelList.remove(*lit);
						delete *lit;
						*lit = NULL;
						continue;
					}
				}

				// we can reconfigure all other labels, even if they are "boxed"
				(*lit)->reconfigure(true);
			}

			// create the rest
			for(StringListIterator sit = labelNames.begin(); sit != labelNames.end(); sit++)
			{
				Label *label = lookupLabel(*sit);
				
				if (!label) 
				{
					label = new Label(*sit);
					labelList.insert(labelList.end(), label);
					
					if ( IsOS(OS_XP) && GetxRCBool((*sit).c_str(), "DropShadow") )
						label->load(hInstance, 0, true);
					else 
						label->load(hInstance);

					//ModuleHook 
					if (!label->moduleHook.empty())
					{
						for(StringListIterator sit = label->moduleHook.begin(); sit != label->moduleHook.end(); sit++)
						{
							string hWndbuffer = IntToString( (int) label->gethWnd() );
							
							string config = (*sit);
							config.append( " \"$themedir$theme.rc\" " );
							config.append( hWndbuffer );

							LSExecute(NULL, config.c_str(), SW_SHOWNORMAL);
						}

						label->hasHookedModules = true;
						label->moduleHook.clear();
					}

					if (!label->isStartHidden())
						label->show();
					else
						label->hide();

					//Hooked Modules cannot Dock!
					if ( label->getBox() != NULL )
						continue;

					char name[64], position[64];
					char *buffers[] = {name, position};
					for(StringListIterator sit = label->dockedToList.begin(); sit != label->dockedToList.end(); sit++)
					{
						LCTokenize((*sit).c_str(), buffers, 2, NULL);
						if ( _stricmp(position, "left") == NULL )
							label->dockedtoSide = 0;
						else if ( _stricmp(position, "top") == NULL )
							label->dockedtoSide = 1;
						else if ( _stricmp(position, "right") == NULL )
							label->dockedtoSide = 2;
						else if ( _stricmp(position, "bottom") == NULL )
							label->dockedtoSide = 3;
						else if ( _stricmp(position, "topleft") == NULL )
							label->dockedtoSide = 4;
						else if ( _stricmp(position, "topright") == NULL )
							label->dockedtoSide = 5;
						else if ( _stricmp(position, "bottomleft") == NULL )
							label->dockedtoSide = 6;
						else if ( _stricmp(position, "bottomright") == NULL )
							label->dockedtoSide = 7;
						else
							label->dockedtoSide = -1;

						bool found = false;
						for(LabelListIterator plit = labelList.begin(); plit != labelList.end(); plit++)
						{
							if( _stricmp(name, (*plit)->getName().c_str()) == NULL )
							{
								(*plit)->hasDockedLabelList.insert((*plit)->hasDockedLabelList.end(), label);
								found = true;
								break;
							}
						}

						if (!found)
						{
							for(LabelListIterator plit = labelList.begin(); plit != labelList.end(); plit++)
							{
								if( _stricmp(name, (*plit)->getLabelGroup().c_str()) == NULL )
								{
									(*plit)->hasDockedLabelList.insert((*plit)->hasDockedLabelList.end(), label);
								}
							}
						}
					}
				}
			}

			return 0;
		}

		case LM_UPDATEBG:
		{
			PaintDesktopEx(0, 0, 0, 0, 0, 0, 0, TRUE);

			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if ( (*lit)->getBox() == 0 && (*lit)->getTransparencyMode() == 2 )
					(*lit)->repaint(true);
			}

			return 0;
		}

		case WM_DISPLAYCHANGE:
		case WM_SETTINGCHANGE:
		{
			PostMessage(hWnd, LM_UPDATEBG, 0, 0);
			return 0;
		}

		case WM_TIMER:
		{
			switch( (int) wParam )
			{
				case TIMER_PLAYSOUND:
				{
					if (m_hWndSoundFile != NULL)
					{
						if (MCIWndGetEnd(m_hWndSoundFile) == MCIWndGetPosition(m_hWndSoundFile))
						{
							MCIWndStop(m_hWndSoundFile);

							KillTimer(messageHandler, TIMER_PLAYSOUND);
						}
					}
					else
					{
						KillTimer(messageHandler, TIMER_PLAYSOUND);
					}
					break;
				}

				default:
				{
					break;
				}
			}
		}
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

BOOL __stdcall _DllMainCRTStartup(HINSTANCE hInst, DWORD fdwReason, LPVOID lpvRes)
{
	// We don't need thread notifications for what we're doing.  Thus, get
	// rid of them, thereby eliminating some of the overhead of this DLL
	DisableThreadLibraryCalls(hInst);

	return TRUE;
}

int initModuleEx(HWND hParent, HINSTANCE hInstance, const char *lsPath)
{
	WNDCLASSEX wc;

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS | CS_DBLCLKS ;
	wc.lpfnWndProc = Label::windowProcedure;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = sizeof(Label *);
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = LoadCursor(0, IDC_ARROW);
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xLabel";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	if(IsOS(OS_XP))
	{
		wc.cbSize = sizeof(WNDCLASSEX);
		wc.style = CS_GLOBALCLASS | CS_DBLCLKS | CS_DROPSHADOW;
		wc.lpfnWndProc = Label::windowProcedure;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = sizeof(Label *);
		wc.hInstance = hInstance;
		wc.hbrBackground = 0;
		wc.hCursor = LoadCursor(0, IDC_ARROW);
		wc.hIcon = 0;
		wc.lpszMenuName = 0;
		wc.lpszClassName = "xLabelXP";
		wc.hIconSm = 0;

		RegisterClassEx(&wc);
	}

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS;
	wc.lpfnWndProc = MessageHandlerProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = LoadCursor(0, IDC_ARROW);
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xLabelMessageHandler";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	messageHandler = CreateWindowEx(WS_EX_TOOLWINDOW,
		"xLabelMessageHandler",
		0,
		WS_POPUP,
		0, 0, 0, 0, 
		0,
		0,
		hInstance,
		0);

	if (!messageHandler)
		return 1;
	
	SendMessage(GetLitestepWnd(),
		LM_REGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	::hInstance = hInstance;

	if ( !IsOS( OS_95 ) && !IsOS( OS_NT4 ) )
	{
		getMonitorInfo();
	
		if ( realMonitors > 1 )
		{
			maxscreenWidth = GetSystemMetrics(SM_CXVIRTUALSCREEN);
			maxscreenHeight = GetSystemMetrics(SM_CYVIRTUALSCREEN);
			screenWidth = GetSystemMetrics(SM_CXSCREEN);
			screenHeight = GetSystemMetrics(SM_CYSCREEN);
		}
		else
		{
			minscreenX = 0;
			minscreenY = 0;
			screenWidth = maxscreenWidth = GetSystemMetrics(SM_CXSCREEN);
			screenHeight = maxscreenHeight = GetSystemMetrics(SM_CYSCREEN);
		}
	}
	else
	{
		minscreenX = 0;
		minscreenY = 0;
		screenWidth = maxscreenWidth = GetSystemMetrics(SM_CXSCREEN);
		screenHeight = maxscreenHeight = GetSystemMetrics(SM_CYSCREEN);
	}

	if (!_PrivateExtractIcons)
	{
		// Bind the icon extraction function
		_PrivateExtractIcons = (UINT (WINAPI *)(LPCTSTR,int,int,int,HICON*,UINT*,UINT,UINT)) GetProcAddress(GetModuleHandle("USER32.DLL"), "PrivateExtractIconsA");
	}

	systemInfo = new SystemInfo();

	AddBangCommand("!LabelAlwaysOnTop", BangAlwaysOnTop);
	AddBangCommand("!LabelClipboardCopy", BangClipboardCopy);
	AddBangCommand("!LabelClipboardPaste", BangClipboardPaste);
	AddBangCommand("!LabelCreate", BangCreate);
	AddBangCommand("!LabelDestroy", BangDestroy);
	AddBangCommand("!LabelDock", BangDock);
	AddBangCommand("!LabelFocus", BangFocus);
	AddBangCommand("!LabelHide", BangHide);
	AddBangCommand("!LabelLSBoxHook", BangLSBoxHook);
	AddBangCommand("!LabelModuleHook", BangModuleHook);
	AddBangCommand("!LabelMove", BangMove);
	AddBangCommand("!LabelMoveBy", BangMoveBy);
	AddBangCommand("!LabelNext", BangNext);
	AddBangCommand("!LabelPrevious", BangPrevious);
	AddBangCommand("!LabelRefresh", BangRefresh);
	AddBangCommand("!LabelReposition", BangReposition);
	AddBangCommand("!LabelRepositionBy", BangRepositionBy);
	AddBangCommand("!LabelResize", BangResize);
	AddBangCommand("!LabelResizeBy", BangResizeBy);
	AddBangCommand("!LabelScroll", BangScroll);
	if ( IsOS(OS_2KXP) )
	{
		AddBangCommand("!LabelSetAlpha", BangSetAlpha);
		AddBangCommand("!LabelGhosted", BangGhosted);
	}
	AddBangCommand("!LabelSetAnimation", BangSetAnimation);
	AddBangCommand("!LabelSetText", BangSetText);
	AddBangCommand("!LabelShow", BangShow);
	AddBangCommand("!LabelShowHide", BangShowHide);
	AddBangCommand("!LabelToggle", BangToggle);
	AddBangCommand("!LabelUpdate", BangUpdate);

	AddBangCommand("!LabelInfoExport", BangInfoExport);

	AddBangCommand("!ParseEvars", BangParseEvars);
	AddBangCommand("!SetEvar", BangSetEvars);

	AddBangCommand("!PlaySound", BangPlaySound);

	AddBangCommand("!TexteditBoxExecute", BangTexteditBoxExecute);

	StringList labelNames = GetxRCNameList("*label");

	for (StringListIterator sit = labelNames.begin(); sit != labelNames.end(); sit++)
	{
		if ( lookupLabel((*sit)) != NULL )
		{
			continue;
		}

		Label *label = new Label(*sit);
		labelList.insert(labelList.end(), label);
		
		if ( IsOS(OS_XP) && Getx2RCBool((*sit), GetxRCString((*sit), "AddToGroup", ""), "DropShadow") )
			label->load(hInstance, 0, true);
		else 
			label->load(hInstance);
		
		//ModuleHook
		if (!label->moduleHook.empty())
		{
			for(StringListIterator sit = label->moduleHook.begin(); sit != label->moduleHook.end(); sit++)
			{
				string hWndbuffer = IntToString( (int) label->gethWnd() );
				
				string config = (*sit);
				config.append( " \"$themedir$theme.rc\" " );
				config.append( hWndbuffer );

				LSExecute(NULL, config.c_str(), SW_SHOWNORMAL);
			}

			label->hasHookedModules = true;
			label->moduleHook.clear();
		}
		
		//Initial Show
		if (!label->isStartHidden())
			label->show();
		else
			label->hide();
	}

	//Docking Setup
	for(LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
	{
		//Hooked Modules cannot Dock!
		if ( (*lit)->getBox() != NULL )
			continue;

		char name[64], position[64];
		char *buffers[] = {name, position};
		for(StringListIterator sit = (*lit)->dockedToList.begin(); sit != (*lit)->dockedToList.end(); sit++)
		{
			LCTokenize((*sit).c_str(), buffers, 2, NULL);
			if ( _stricmp(position, "left") == NULL )
				(*lit)->dockedtoSide = 0;
			else if ( _stricmp(position, "top") == NULL )
				(*lit)->dockedtoSide = 1;
			else if ( _stricmp(position, "right") == NULL )
				(*lit)->dockedtoSide = 2;
			else if ( _stricmp(position, "bottom") == NULL )
				(*lit)->dockedtoSide = 3;
			else if ( _stricmp(position, "topleft") == NULL )
				(*lit)->dockedtoSide = 4;
			else if ( _stricmp(position, "topright") == NULL )
				(*lit)->dockedtoSide = 5;
			else if ( _stricmp(position, "bottomleft") == NULL )
				(*lit)->dockedtoSide = 6;
			else if ( _stricmp(position, "bottomright") == NULL )
				(*lit)->dockedtoSide = 7;
			else
				(*lit)->dockedtoSide = -1;

			bool found = false;
			for(LabelListIterator plit = labelList.begin(); plit != labelList.end(); plit++)
			{
				if( _stricmp( name, (*plit)->getName().c_str()) == NULL )
				{
					(*plit)->hasDockedLabelList.insert((*plit)->hasDockedLabelList.end(), (*lit));
					found = true;
					break;
				}
			}

			if (!found)
			{
				for(LabelListIterator plit = labelList.begin(); plit != labelList.end(); plit++)
				{
					if( _stricmp(name, (*plit)->getLabelGroup().c_str()) == NULL )
					{
						(*plit)->hasDockedLabelList.insert((*plit)->hasDockedLabelList.end(), (*lit));
					}
				}
			}
		}
	}

	return 0;
}

extern HDC hdcDesktop;
extern HBITMAP hbmDesktop;

void quitModule(HINSTANCE hInstance)
{
	if (m_hWndSoundFile != NULL)
	{
		MCIWndStop(m_hWndSoundFile);
		MCIWndDestroy(m_hWndSoundFile);
		m_hWndSoundFile = NULL;
	}

	RemoveBangCommand("!LabelAlwaysOnTop");
	RemoveBangCommand("!LabelClipboardCopy");
	RemoveBangCommand("!LabelClipboardPaste");
	RemoveBangCommand("!LabelCreate");
	RemoveBangCommand("!LabelDestroy");
	RemoveBangCommand("!LabelDock");
	RemoveBangCommand("!LabelFocus");
	RemoveBangCommand("!LabelHide");
	RemoveBangCommand("!LabelLSBoxHook");
	RemoveBangCommand("!LabelModuleHook");
	RemoveBangCommand("!LabelMove");
	RemoveBangCommand("!LabelMoveBy");
	RemoveBangCommand("!LabelNext");
	RemoveBangCommand("!LabelPrevious");
	RemoveBangCommand("!LabelRefresh");
	RemoveBangCommand("!LabelReposition");
	RemoveBangCommand("!LabelRepositionBy");
	RemoveBangCommand("!LabelResize");
	RemoveBangCommand("!LabelResizeBy");
	RemoveBangCommand("!LabelScroll");
	if ( IsOS(OS_2KXP) )
	{
		RemoveBangCommand("!LabelSetAlpha");
		RemoveBangCommand("!LabelGhosted");
	}
	RemoveBangCommand("!LabelSetAnimation");
	RemoveBangCommand("!LabelSetText");
	RemoveBangCommand("!LabelShow");
	RemoveBangCommand("!LabelShowHide");
	RemoveBangCommand("!LabelToggle");
	RemoveBangCommand("!LabelUpdate");

	RemoveBangCommand("!LabelInfoExport");

	RemoveBangCommand("!ParseEvars");
	RemoveBangCommand("!SetEvar");

	RemoveBangCommand("!PlaySound");

	RemoveBangCommand("!TexteditBoxExecute");

	for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
	{
		(*lit)->hide();
		delete *lit;
		*lit = NULL;
	}

	labelList.clear();

	SendMessage(GetLitestepWnd(),
		LM_UNREGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	DestroyWindow(messageHandler);

	UnregisterClass("xLabel", hInstance);
	if(IsOS(OS_XP))
		UnregisterClass("xLabelXP", hInstance);
	UnregisterClass("xLabelMessageHandler", hInstance);

	if (systemInfo)
		delete systemInfo;
	
	if (hbmDesktop != 0 || hdcDesktop != 0)
	{
		hbmDesktop = (HBITMAP) SelectObject(hdcDesktop, hbmDesktop);
		DeleteObject(hbmDesktop);
		DeleteDC(hdcDesktop);
	}
}

//Text EditBox
//-----------------------------------------------------
//LRESULT WINAPI TextEditProc(HWND hText, UINT msg, WPARAM wParam, LPARAM lParam);
LRESULT WINAPI TextEditProc(HWND hText, UINT msg, WPARAM wParam, LPARAM lParam)
{
	Label* label = (Label *) GetWindowLong(GetParent(hText), 0);

	char buf[MAX_PATH] = {0};

	switch(msg)
	{
		case WM_CHAR:
		{
			if( wParam == VK_RETURN )
			{
				GetWindowText(hText, buf, sizeof(buf));

				SendMessage(label->gethWnd(), LM_TEXTEDITBOX, 0, (LONG) buf);
			}
			else if ( wParam == VK_ESCAPE )
			{
				ParseBangCommand(label->gethWnd(), "!labelhide", (label->getName() + " .editbox").c_str());
			}
		}
		break;

		case WM_SETCURSOR:
		{
			string process = label->getTextCursorPath();
			if ( process.empty() )
			{
				break;
			}
			else if ( _stricmp(process.c_str(), ".none") == NULL )
			{
				return TRUE;
			}
			else if ( label->getTextCursor() )
			{
				SetCursor( label->getTextCursor() );
				return TRUE;
			}
		
			return FALSE;
		}
		break;

		case WM_SETFOCUS:
		{
			ShowCaret(label->getTexteditBoxhWnd());

			if ( !label->focusCommand.empty() )
				xLSExecute(label->gethWnd(), label->focusCommand.c_str(), SW_SHOWNORMAL);	
		}	
		break;

		case WM_KILLFOCUS:
		{
			if ( !label->unfocusCommand.empty() )
				xLSExecute(label->gethWnd(), label->unfocusCommand.c_str(), SW_SHOWNORMAL);
		}
		break;
    }

	return CallWindowProc(label->wpOrigEditProc, hText, msg, wParam, lParam);
}
//-----------------------------------------------------

Label *lookupLabel( const string &name )
{
	for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
	{
		if (_stricmp(name.c_str(), (*lit)->getName().c_str()) == NULL)
			return *lit;
	}

	return NULL;
}

// !LabelInfoExport < clipboard | %@LSLua var%@ | %%mzscript var%% | %#Evar%# > <Text Escape Sequence>
void BangInfoExport(HWND hwndCaller, LPCSTR pszArgs)
{
	string Bang = "";
	char namevar[64], textvar[MAX_LINE_LENGTH];
	char *buffers[] = {namevar, textvar};

	if (LCTokenize(pszArgs, buffers, 2, NULL) == 2)
	{	
		Label *label = lookupLabel(namevar);

		if(label)
		{
			string evar = textvar;

			if (strstr(evar.c_str(), "%#") != NULL)
			{
				while ( strstr(evar.c_str(), "%#") != NULL )
				{
					evar.replace(evar.find("%#"), 2, "", 0);
				}

				Bang = ("\"" + label->getText() + "\"");
				PrefixedLSSetVariable( evar, "", Bang );
			}
			else if (strstr(evar.c_str(), "%%") != NULL)
			{
				while ( strstr(evar.c_str(), "%%") != NULL )
				{
					evar.replace(evar.find("%%"), 2, "", 0);
				}

				Bang = evar;
				Bang.append(" \"" + label->getText() + "\"");
				ParseBangCommand(NULL, "!varset", Bang.c_str());
			}
			else if (strstr(evar.c_str(), "%@") != NULL)
			{
				while ( strstr(evar.c_str(), "%@") != NULL )
				{
					evar.replace(evar.find("%@"), 2, "", 0);
				}

				CString lslua = label->getText().c_str();
				lslua.Replace("\\", "\\\\");
				lslua.Replace("\"", "\\\"");
				lslua.Replace("'", "\\'");

				/*Bang = "\"";
				Bang.append( evar );
				Bang.append(" = \\\"" + lslua + "\\\"\"");
				ParseBangCommand(NULL, "!luaexec", Bang.c_str());*/

				/*Bang = evar;
				Bang.append(" = \\\"" + lslua + "\\\"");
				ParseBangCommand(NULL, "!luaexec", Bang.c_str());*/

				Bang = evar;
				Bang.append(" = '" + lslua + "'");
				ParseBangCommand(NULL, "!luaexec", Bang.c_str());

				//!luaexec "myevar = \"This is the Lua String\""
			}
			else
			{
				SetClipboardText(NULL, label->getText().c_str());
			}
		}
		else
		{
			bool dynamic = false;
			string escaperesult = "";
			string evar = namevar;

			if (strstr(evar.c_str(), "%#") != NULL)
			{
				while ( strstr(evar.c_str(), "%#") != NULL )
				{
					evar.replace(evar.find("%#"), 2, "", 0);
				}

				escaperesult = systemInfo->processLabelText(textvar, (*labelList.begin()), &dynamic);

				Bang = ("\"" + escaperesult + "\"");
				PrefixedLSSetVariable( evar, "", Bang );
			}
			else if (strstr(evar.c_str(), "%%") != NULL)
			{
				while ( strstr(evar.c_str(), "%%") != NULL )
				{
					evar.replace(evar.find("%%"), 2, "", 0);
				}

				escaperesult = systemInfo->processLabelText(textvar, (*labelList.begin()), &dynamic);

				Bang = evar;
				Bang.append(" \"" + escaperesult + "\"");
				ParseBangCommand(NULL, "!varset", Bang.c_str());
			}
			else if (strstr(evar.c_str(), "%@") != NULL)
			{
				while ( strstr(evar.c_str(), "%@") != NULL )
				{
					evar.replace(evar.find("%@"), 2, "", 0);
				}

				escaperesult = systemInfo->processLabelText(textvar, (*labelList.begin()), &dynamic);

				CString lslua = escaperesult.c_str();
				lslua.Replace("\\", "\\\\");
				lslua.Replace("\"", "\\\"");
				lslua.Replace("'", "\\'");

				/*Bang = "\"";
				Bang.append( evar );
				Bang.append(" = \\\"" + lslua + "\\\"\"");
				ParseBangCommand(NULL, "!luaexec", Bang.c_str());*/

				/*Bang = evar;
				Bang.append(" = \\\"" + lslua + "\\\"");
				ParseBangCommand(NULL, "!luaexec", Bang.c_str());*/

				Bang = evar;
				Bang.append(" = '" + lslua + "'");
				ParseBangCommand(NULL, "!luaexec", Bang.c_str());

				//!luaexec "myevar = \"This is the Lua String\""
			}
			else
			{
				if ( _stricmp(namevar, "clipboard") == NULL )
					escaperesult = systemInfo->processLabelText(textvar, (*labelList.begin()), &dynamic);
				else
					escaperesult = systemInfo->processLabelText(namevar, (*labelList.begin()), &dynamic);

				SetClipboardText(NULL, escaperesult.c_str());
			}
		}
	}

}

// !LabelSetAnimation <label> <on | off> [<loops>]
void BangSetAnimation(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], option[16], loops[16];
	char *buffers[] = {name, option, loops};

	int numTokens = LCTokenize(pszArgs, buffers, 3, NULL);
	if (numTokens >= 2)
	{
		Label *label = lookupLabel(name);

		if (label)
		{
			if (numTokens == 2)
			{
				if (_stricmp(option, "false") == NULL || _stricmp(option, "off") == NULL)
					label->setAnimation(false);
				else if (_stricmp(option, "true") == NULL || _stricmp(option, "on") == NULL)
					label->setAnimation(true);
			}
			else if (numTokens == 3)
			{
				if (_stricmp(option, "false") == NULL || _stricmp(option, "off") == NULL)
					label->setAnimation(false);
				else if (_stricmp(option, "true") == NULL || _stricmp(option, "on") == NULL)
					label->setAnimation(true, atoi(loops));
			}
		}
	}
}

// !LabelAlwaysOnTop <label> <on|true, off|false, toggle>
void BangAlwaysOnTop(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], option[16];
	char *buffers[] = {name, option};

	LCTokenize(pszArgs, buffers, 2, NULL);

	Label *label = lookupLabel(name);
	if(label)
	{
		if (_stricmp(option, "false") == NULL || _stricmp(option, "off") == NULL)
			label->setAlwaysOnTop(false);
		else if (_stricmp(option, "toggle") == NULL || _stricmp(option, "switch") == NULL)
			label->setAlwaysOnTop(!label->isAlwaysOnTop());
		else
			label->setAlwaysOnTop(true);
	}
	else
	{
		for(LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++ )
		{
			if( _stricmp((*lit)->labelGroup.c_str(), name) == NULL )
			{
				if (_stricmp(option, "false") == NULL || _stricmp(option, "off") == NULL)
					(*lit)->setAlwaysOnTop(false);
				else if (_stricmp(option, "toggle") == NULL || _stricmp(option, "switch") == NULL)
					(*lit)->setAlwaysOnTop(!(*lit)->isAlwaysOnTop());
				else
					(*lit)->setAlwaysOnTop(true);
			}
		}
	}
}

// !LabelClipboardCopy <label> <prefix>
void BangClipboardCopy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], prefix[256];
	char *buffers[] = {name, prefix};
	int numTokens = LCTokenize(pszArgs, buffers, 2, NULL);

	if(numTokens >= 1)
	{
		Label *label = lookupLabel(name);

		if(label)
			label->clipboardCopy((numTokens >= 2) ? prefix : "");
	}
}

// !LabelClipboardPaste <label> <prefix>
void BangClipboardPaste(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], prefix[256];
	char *buffers[] = {name, prefix};
	int numTokens = LCTokenize(pszArgs, buffers, 2, NULL);

	if(numTokens >= 1)
	{
		Label *label = lookupLabel(name);

		if(label)
			label->clipboardPaste((numTokens >= 2) ? prefix : "");
	}
}

// !LabelCreate <label>
void BangCreate(HWND hwndCaller, LPCSTR pszArgs)
{
	if ( lookupLabel( pszArgs ) != NULL )
		return;

	Label *label = new Label(pszArgs);

	labelList.insert(labelList.end(), label);

	if ( IsOS(OS_XP) && GetxRCBool(pszArgs, "DropShadow") )
		label->load(hInstance, 0, true);
	else 
		label->load(hInstance);

	//ModuleHook
	if (!label->moduleHook.empty())
	{
		for(StringListIterator sit = label->moduleHook.begin(); sit != label->moduleHook.end(); sit++)
		{
			string hWndbuffer = IntToString( (int) label->gethWnd() );
			
			string config = (*sit);
			config.append( " \"$themedir$theme.rc\" " );
			config.append( hWndbuffer );

			LSExecute(NULL, config.c_str(), SW_SHOWNORMAL);
		}

		label->hasHookedModules = true;
		label->moduleHook.clear();
	}

	if (!label->isStartHidden())
		label->show();
	else
		label->hide();

	//Docking
	char name[64], position[64];
	char *buffers[] = {name, position};
	for(StringListIterator sit = label->dockedToList.begin(); sit != label->dockedToList.end(); sit++)
	{
		LCTokenize((*sit).c_str(), buffers, 2, NULL);
		if ( _stricmp(position, "left") == NULL )
			label->dockedtoSide = 0;
		else if ( _stricmp(position, "top") == NULL )
			label->dockedtoSide = 1;
		else if ( _stricmp(position, "right") == NULL )
			label->dockedtoSide = 2;
		else if ( _stricmp(position, "bottom") == NULL )
			label->dockedtoSide = 3;
		else if ( _stricmp(position, "topleft") == NULL )
			label->dockedtoSide = 4;
		else if ( _stricmp(position, "topright") == NULL )
			label->dockedtoSide = 5;
		else if ( _stricmp(position, "bottomleft") == NULL )
			label->dockedtoSide = 6;
		else if ( _stricmp(position, "bottomright") == NULL )
			label->dockedtoSide = 7;
		else
			label->dockedtoSide = -1;

		bool found = false;
		for(LabelListIterator plit = labelList.begin(); plit != labelList.end(); plit++)
		{
			if( _stricmp( (*sit).c_str(), (*plit)->getName().c_str()) == NULL )
			{
				(*plit)->hasDockedLabelList.insert((*plit)->hasDockedLabelList.end(), label);
				found = true;
				break;
			}
		}

		if (!found)
		{
			for(LabelListIterator plit = labelList.begin(); plit != labelList.end(); plit++)
			{
				if( _stricmp( (*sit).c_str(), (*plit)->getLabelGroup().c_str()) == NULL )
				{
					(*plit)->hasDockedLabelList.insert((*plit)->hasDockedLabelList.end(), label);
				}
			}
		}
	}
}

// !LabelDestroy <label>
void BangDestroy(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = lookupLabel(pszArgs);
	
	if (label)
	{
		label->hide();
		labelList.remove(label);
		delete label;
	}
	else
	{
		bool rescan = true;
		bool skip = false;
		while ( rescan )
		{
			skip = false;
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if( _stricmp((*lit)->labelGroup.c_str(), pszArgs) == NULL )
				{
					(*lit)->hide();
					labelList.remove((*lit));
					delete (*lit);

					skip = true;
					break;
				}
			}

			if ( !skip )
				rescan = false;
		}
	}
}

// !LabelDock <label> <label> <"true", "false"> <modifier>
void BangDock(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], dockto[64], state[32], modifier[32];
	char *buffers[] = {name, dockto, state, modifier};

	if(LCTokenize(pszArgs, buffers, 4, NULL) >= 2)
	{	
		Label *label = lookupLabel(name);
		Label *labeldockto = lookupLabel(dockto);

		if (_stricmp( state, "false" ) == NULL)
		{
			if(label && labeldockto)
			{
				for(LabelListIterator lit = labeldockto->hasDockedLabelList.begin(); lit != labeldockto->hasDockedLabelList.end(); lit++)
				{
					if ( _stricmp( name, (*lit)->getName().c_str() ) == NULL )
					{
						labeldockto->hasDockedLabelList.erase ( lit );
						break;
					}
				}	
			}
		}
		else
		{
			if(label && labeldockto)
			{
				bool alreadyFound = false;
				for(LabelListIterator lit = labeldockto->hasDockedLabelList.begin(); lit != labeldockto->hasDockedLabelList.end(); lit++)
				{
					if ( _stricmp( name, (*lit)->getName().c_str() ) == NULL )
					{
						alreadyFound = true;
						break;
					}
				}
				if ( !alreadyFound )
				{
					if ( _stricmp(modifier, "left") == NULL )
						label->dockedtoSide = 0;
					else if ( _stricmp(modifier, "top") == NULL )
						label->dockedtoSide = 1;
					else if ( _stricmp(modifier, "right") == NULL )
						label->dockedtoSide = 2;
					else if ( _stricmp(modifier, "bottom") == NULL )
						label->dockedtoSide = 3;
					else if ( _stricmp(modifier, "topleft") == NULL )
						label->dockedtoSide = 4;
					else if ( _stricmp(modifier, "topright") == NULL )
						label->dockedtoSide = 5;
					else if ( _stricmp(modifier, "bottomleft") == NULL )
						label->dockedtoSide = 6;
					else if ( _stricmp(modifier, "bottomright") == NULL )
						label->dockedtoSide = 7;
					else
						label->dockedtoSide = -1;

					labeldockto->hasDockedLabelList.insert ( labeldockto->hasDockedLabelList.end(), label );
				}
			}
		}
	}
}

// !LabelFocus <label>
void BangFocus(HWND hwndCaller, LPCSTR pszArgs)
{
	bool textEditBox = false;
	char name[64] = {0};
	char optional[64] = {0};
	char *buffers[] = {name, optional};

	LCTokenize(pszArgs, buffers, 2, NULL);

	if ( _stricmp(optional, ".editbox") == NULL )
		textEditBox = true;

	Label *label = lookupLabel(name);

	if ( label )
	{	
		if ( !textEditBox )
		{
			SetForegroundWindow(label->gethWnd());
			SetFocus(label->gethWnd());
		}
		else
		{
			string bang = "!LabelShow ";
			bang.append( label->getName() );
			bang.append( " .editbox" );
			LSExecute(NULL, bang.c_str(), SW_SHOWNORMAL);
		}
	}
}

// !LabelShow <label> [<overlaylabel> | ".editbox"]
void BangShow(HWND hwndCaller, LPCSTR pszArgs)
{
	bool textEditBox = false;
	char name[64] = {0};
	char optional[64] = {0};
	char *buffers[] = {name, optional};

	int numTokens = LCTokenize(pszArgs, buffers, 2, NULL);

	if ( _stricmp(optional, ".editbox") == NULL )
		textEditBox = true;

	Label *label = lookupLabel(name);

	if (label)
	{
		if ( numTokens >= 2 && !textEditBox )
		{
			for(GhostLabelListIterator glit = label->ghostlabelList.begin(); glit != label->ghostlabelList.end(); glit++)
			{ 
				if ( *glit != NULL )
				{
					if ( _stricmp((*glit)->name.c_str(), optional) == NULL || _stricmp((*glit)->labelGroup.c_str(), optional) == NULL )
					{
						label->show(false, 2, (*glit));
					}
				}
				else
				{
					break;
				}
			}
		}
		else if ( textEditBox )
			label->show(false, 1);
		else
			label->show(true);
	}
	else
	{
		for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
		{
			if( _stricmp((*lit)->labelGroup.c_str(), name) == NULL )
			{
				if ( numTokens >= 2 && !textEditBox )
				{
					for(GhostLabelListIterator glit = (*lit)->ghostlabelList.begin(); glit != (*lit)->ghostlabelList.end(); glit++)
					{ 
						if ( *glit != NULL )
						{
							if ( _stricmp((*glit)->name.c_str(), optional) == NULL || _stricmp((*glit)->labelGroup.c_str(), optional) == NULL )
							{
								(*lit)->show(false, 2, (*glit));
							}
						}
						else
						{
							break;
						}
					}
				}
				else if ( textEditBox )
					(*lit)->show(false, 1);
				else
					(*lit)->show(true);
			}
		}
	}
}

// !LabelShowHide <label> <timeout>
void BangShowHide(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[64];
	char *buffers[] = {name, A};

	if (LCTokenize(pszArgs, buffers, 2, NULL) >= 2)
	{
		int timeout = atoi(A);

		Label *label = lookupLabel(name);

		if (label)
		{
			label->showHide(timeout, true);
		}
		else
		{
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if ( _stricmp((*lit)->labelGroup.c_str(), name) == NULL )
					(*lit)->showHide(timeout, true);
			}
		}
	}
}

// !LabelHide <label> [<overlaylabel> | ".editbox"]
void BangHide(HWND hwndCaller, LPCSTR pszArgs)
{
	bool textEditBox = false;
	char name[64] = {0};
	char optional[64] = {0};
	char *buffers[] = {name, optional};

	int numTokens = LCTokenize(pszArgs, buffers, 2, NULL);

	if ( _stricmp(optional, ".editbox") == NULL )
		textEditBox = true;

	Label *label = lookupLabel(name);
	
	if (label)
	{
		if ( numTokens >= 2 && !textEditBox )
		{
			for(GhostLabelListIterator glit = label->ghostlabelList.begin(); glit != label->ghostlabelList.end(); glit++)
			{ 
				if ( *glit != NULL )
				{
					if ( _stricmp((*glit)->name.c_str(), optional) == NULL || _stricmp((*glit)->labelGroup.c_str(), optional) == NULL )
					{
						label->hide(false, 2, (*glit));
					}
				}
				else
				{
					break;
				}
			}
		}
		else if ( textEditBox )
			label->hide(false, 1);
		else
			label->hide(true);
	}
	else
	{
		for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++ )
		{
			if( _stricmp((*lit)->labelGroup.c_str(), name) == NULL )
			{
				if ( numTokens >= 2 && !textEditBox )
				{
					for(GhostLabelListIterator glit = (*lit)->ghostlabelList.begin(); glit != (*lit)->ghostlabelList.end(); glit++)
					{ 
						if ( *glit != NULL )
						{
							if ( _stricmp((*glit)->name.c_str(), optional) == NULL )
							{
								(*lit)->hide(false, 2, (*glit));
							}
						}
						else
						{
							break;
						}
					}
				}
				else if ( textEditBox )
					(*lit)->hide(false, 1);
				else
					(*lit)->hide(true);
			}
		}
	}
}

// !LabelModuleHook <labelname> <hooking bang>
void BangModuleHook(HWND hwndCaller, LPCSTR pszArgs)
{
	char nameparent[64], rest[MAX_LINE_LENGTH];
	char *buffers[] = {nameparent};

	LCTokenize(pszArgs, buffers, 1, rest);

	Label *label = lookupLabel(nameparent);

	if (label)
	{
		if ( !label->hasCorrectAlphaState() )
		{
			ModifyStyleEx(label->gethWnd(), WS_EX_LAYERED, 0);
		}

		string hWndbuffer = IntToString( (int) label->gethWnd() );
					
		string config = rest;
		config.append( " \"$themedir$theme.rc\" " );
		config.append( hWndbuffer );

		LSExecute(NULL, config.c_str(), SW_SHOWNORMAL);

		if ( !label->hasCorrectAlphaState() )
		{	
			label->hasHookedModules = true;
			label->setFakeAlphaMap(false);

			ModifyStyle(label->gethWnd(), WS_POPUP, WS_CHILD);
			SetParent(label->gethWnd(), 0);
			ModifyStyle(label->gethWnd(), WS_CHILD, WS_POPUP);

			ModifyStyleEx(label->gethWnd(), 0, WS_EX_LAYERED);

			ModifyStyle(label->gethWnd(), WS_POPUP, WS_CHILD);
			SetParent(label->gethWnd(), label->isAlwaysOnTop() ? 0 : GetLitestepDesktop());
			ModifyStyle(label->gethWnd(), WS_CHILD, WS_POPUP);

			label->setAlpha(255, true);
		}
	}
}

// !LabelLSBoxHook <labelname>
void BangLSBoxHook(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], rest[MAX_LINE_LENGTH];
	char *buffers[] = {name};

	LCTokenize(pszArgs, buffers, 1, rest);

	const char *handle = strrchr(pszArgs,' ');

	HWND hwndBox = (HWND)atoi(handle+1);

	Label *label = lookupLabel(name);

	if (label)
	{
		label->hide();
		labelList.remove(label);
		delete label;	
	}

	label = new Label(name);
	labelList.insert(labelList.end(), label);

	if ( IsOS(OS_XP) && GetxRCBool(name, "DropShadow") )
		label->load(hInstance, hwndBox, true);
	else 
		label->load(hInstance, hwndBox);

	//ModuleHook
	if (!label->moduleHook.empty())
	{
		for(StringListIterator sit = label->moduleHook.begin(); sit != label->moduleHook.end(); sit++)
		{
			string hWndbuffer = IntToString( (int) label->gethWnd() );
				
			string config = (*sit);
			config.append( " \"$themedir$theme.rc\" " );
			config.append( hWndbuffer );

			LSExecute(NULL, config.c_str(), SW_SHOWNORMAL);
		}

		label->hasHookedModules = true;
		label->moduleHook.clear();
	}

	if (!label->isStartHidden())
		label->show();
	else
		label->hide();
}

// !LabelMove <label> <x> <y> [<steps> <time>]
void BangMove(HWND hwndCaller, LPCSTR pszArgs)
{
	int x = 0;
	int y = 0;
	char name[64], A[16], B[16], C[16], D[16];
	char *buffers[] = {name, A, B, C, D};

	if (LCTokenize(pszArgs, buffers, 5, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		if (label)
		{
			if (label->getBox() == NULL)
			{
				x = xParseCoord(A, 0, screenWidth, NULL, false);
				y = xParseCoord(B, 0, screenHeight, NULL, true);
			}
			else
			{
				RECT rc;
				GetClientRect(label->getBox(), &rc);
				x = xParseCoord(A, 0, rc.right, label->getBox());
				y = xParseCoord(B, 0, rc.bottom, label->getBox());
			}
			int step = atoi( C );
			int time = atoi( D );

			label->reposition(1, x, y, 0, 0, step, time);
		}
	}
}

// !LabelMoveBy <label> <x> <y> [<steps> <time>] [bool s]
void BangMoveBy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16], C[16], D[16];
	char *buffers[] = {name, A, B, C, D};

	if (LCTokenize(pszArgs, buffers, 5, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		int x = atoi( A );
		int y = atoi( B );

		int step = atoi( C );
		int time = atoi( D );

		if (label)
		{
			label->reposition(1, x, y, 0, 0, step, time, true);
		}
		else
		{
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if ( _stricmp((*lit)->labelGroup.c_str(), name) == NULL )
					(*lit)->reposition(1, x, y, 0, 0, step, time, true);
			}
		}
	}
}

// !LabelNext <label>
void BangNext(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = lookupLabel(pszArgs);

	if (label)
		label->next();
}

// !LabelPrevious <label>
void BangPrevious(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = lookupLabel(pszArgs);

	if (label)
		label->previous();
}

// !LabelRefresh <label>
void BangRefresh(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], setting[256], value[MAX_LINE_LENGTH];
	char *buffers[] = {name, setting};

	int tokens = LCTokenize(pszArgs, buffers, 2, value);

	Label *label = lookupLabel(name);

	if (label)
	{
		if (tokens == 1)
		{
			label->reconfigure(true);
		}

		if (tokens == 2)
		{
			PrefixedLSSetVariable( (string)setting, "", (string)value );
			label->reconfigure(true);
		}
	}
	else
	{
		for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
		{
			if ( _stricmp((*lit)->labelGroup.c_str(), name) == NULL )
			{
				if (tokens == 1)
				{
					label->reconfigure(true);
				}

				if (tokens == 2)
				{
					string worker = setting;
					if (strstr(setting, "*") != NULL)
						worker.replace(worker.find("*"), 1, (*lit)->getName().c_str(), (*lit)->getName().length());

					PrefixedLSSetVariable( worker, "", (string)value );
					(*lit)->reconfigure(true);
				}
			}
		}
	}
}

// !LabelReposition <label> <x> <y> <width> <height> [<steps>] [<time>]
void BangReposition(HWND hwndCaller, LPCSTR pszArgs)
{
	int x = 0;
	int y = 0;
	int w = 64;
	int h = 64;
	char name[64], A[16], B[16], C[16], D[16], E[16], F[16];
	char *buffers[] = {name, A, B, C, D, E, F};

	if(LCTokenize(pszArgs, buffers, 7, NULL) >= 5)
	{
		Label *label = lookupLabel(name);

		if (label)
		{
			if (label->getBox() == NULL)
			{
				x = xParseCoord(A, 0, screenWidth, NULL, false);
				y = xParseCoord(B, 0, screenHeight, NULL, true);
				w = xParseDimen(C, 0, screenWidth, NULL, false);
				h = xParseDimen(D, 0, screenHeight, NULL, true);
			}
			else
			{
				RECT rc;
				GetClientRect(label->getBox(), &rc);
				x = xParseCoord(A, 0, rc.right, label->getBox());
				y = xParseCoord(B, 0, rc.bottom, label->getBox());
				w = xParseDimen(C, 0, rc.right, label->getBox());
				h = xParseDimen(D, 0, rc.bottom, label->getBox());
			}
			int step = atoi( E );
			int time = atoi( F );

			label->reposition(3, x, y, w, h, step, time);
		}
	}
}

// !LabelRepositionBy <label> <cx> <cy> <cwidth> <cheight> [<steps>] [<time>]
void BangRepositionBy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16], C[16], D[16], E[16], F[16];
	char *buffers[] = {name, A, B, C, D, E, F};

	if (LCTokenize(pszArgs, buffers, 7, NULL) >= 5)
	{
		Label *label = lookupLabel(name);
		
		int x = atoi( A );
		int y = atoi( B );
		int w = atoi( C );
		int h = atoi( D );

		int step = atoi( E );
		int time = atoi( F );

		if (label)
		{
			label->reposition(3, x, y, w, h, step, time, true);
		}
		else
		{
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if ( _stricmp((*lit)->labelGroup.c_str(), name) == NULL )
					(*lit)->reposition(3, x, y, w, h, step, time, true);
			}
		}
	}
}


// !LabelResize <label> <width> <height> [<steps> <time>]
void BangResize(HWND hwndCaller, LPCSTR pszArgs)
{
	int w = 64;
	int h = 64;
	char name[64], A[16], B[16], C[16], D[16];
	char *buffers[] = {name, A, B, C, D};

	if (LCTokenize(pszArgs, buffers, 5, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		if (label)
		{
			if (label->getBox() == NULL)
			{
				w = xParseDimen(A, 0, screenWidth, NULL, false);
				h = xParseDimen(B, 0, screenHeight, NULL, true);
			}
			else
			{
				RECT rc;
				GetClientRect(label->getBox(), &rc);
				w = xParseDimen(A, 0, rc.right, label->getBox());
				h = xParseDimen(B, 0, rc.bottom, label->getBox());
			}
			int step = atoi( C );
			int time = atoi( D );

			label->reposition(2, 0, 0, w, h, step, time);
		}
	}
}

// !LabelResizeBy <label> <cx> <cy> [<steps> <time>]
void BangResizeBy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16], C[16], D[16];
	char *buffers[] = {name, A, B, C, D};

	if (LCTokenize(pszArgs, buffers, 5, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		int cx = atoi( A );
		int cy = atoi( B );

		int step = atoi( C );
		int time = atoi( D );

		if (label)
		{
			label->reposition(2, 0, 0, cx, cy, step, time, true);
		}
		else
		{
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if ( _stricmp((*lit)->labelGroup.c_str(), name) == NULL )
					(*lit)->reposition(2, 0, 0, cx, cy, step, time, true);
			}
		}
	}
}

// !LabelScroll <label> <'on', 'off', 'toggle', 'freeze', limit> <scrollMode(1-4)>
void BangScroll(HWND hwndCaller, LPCSTR pszArgs)
{
	// FIXME: This should be split into multiple bang commands
	char name[64], option[16], mode[16];
	char *buffers[] = {name, option, mode};

	if (LCTokenize(pszArgs, buffers, 3, NULL) >= 2)
	{
		Label *label = lookupLabel(name);

		if (label)
		{
			if (_stricmp(option, "false") == NULL || _stricmp(option, "off") == NULL)
			{
				label->setScrolling(0);
				if (label->frozen)
					label->frozen = false;
			}
			else if (_stricmp(option, "true") == NULL || _stricmp(option, "on") == NULL)
			{
				int scrollmode = atoi(mode);
				if (scrollmode > 0 && scrollmode < 5)
					label->setScrolling(scrollmode);
				else
					label->setScrolling(label->scrollBackup);

				label->repaint();
			}
			else if (_stricmp(option, "toggle") == NULL || _stricmp(option, "switch") == NULL)
			{
				if (label->getScrolling() > 0)
				{
					label->setScrolling(0);
					if (label->frozen)
						label->frozen = false;
				}
				else
				{
					int scrollmode = atoi(mode);
					if (scrollmode > 0 && scrollmode < 5)
						label->setScrolling(scrollmode);
					else
						label->setScrolling(label->scrollBackup);

					label->repaint();
				}
			}
			else if (_stricmp(option, "freeze") == NULL)
			{
				if (label->frozen)
					label->frozen = false;
				else
					label->frozen = true;
			}
			else
			{
				int limit = atoi(option);
				if (limit >= 0) 
					label->setScrollLimit(limit);
			}
		}
		else
		{
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if ( _stricmp((*lit)->labelGroup.c_str(), name) == NULL )
				{
					if (_stricmp(option, "false") == NULL || _stricmp(option, "off") == NULL)
					{
						(*lit)->setScrolling(0);
						if ((*lit)->frozen)
							(*lit)->frozen = false;
					}
					else if (_stricmp(option, "true") == NULL || _stricmp(option, "on") == NULL)
					{
						int scrollmode = atoi(mode);
						if (scrollmode > 0 && scrollmode < 5)
							(*lit)->setScrolling(scrollmode);
						else
							(*lit)->setScrolling((*lit)->scrollBackup);

						(*lit)->repaint();
					}
					else if (_stricmp(option, "toggle") == NULL || _stricmp(option, "switch") == NULL)
					{
						if ((*lit)->getScrolling() > 0)
						{
							(*lit)->setScrolling(0);
							if ((*lit)->frozen)
								(*lit)->frozen = false;
						}
						else
						{
							int scrollmode = atoi(mode);
							if (scrollmode > 0 && scrollmode < 5)
								(*lit)->setScrolling(scrollmode);
							else
								(*lit)->setScrolling((*lit)->scrollBackup);

							(*lit)->repaint();
						}
					}
					else if (_stricmp(option, "freeze") == NULL)
					{
						if ((*lit)->frozen)
							(*lit)->frozen = false;
						else
							(*lit)->frozen = true;
					}
					else
					{
						int limit = atoi(option);
						if (limit >= 0) 
							(*lit)->setScrollLimit(limit);
					}
				}
			}
		}
	}
}

// !LabelSetText <label> [<overlaylabel> | ".editbox" ] <text>
void BangSetText(HWND hwndCaller, LPCSTR pszArgs)
{
	bool textEditBox = false;
	char name[64] = {0};
	char optional[MAX_LINE_LENGTH] = {0};
	char text[MAX_LINE_LENGTH] = {0};
	char *buffers[] = {name, optional, text};
	int numTokens = LCTokenize(pszArgs, buffers, 3, NULL);

	if ( numTokens == 3 && _stricmp(optional, ".editbox") == NULL )
		textEditBox = true;

	if (numTokens >= 1)
	{
		Label *label = lookupLabel(name);
	
		if (label)
		{
			if ( numTokens == 3 && !textEditBox )
			{
				for(GhostLabelListIterator glit = label->ghostlabelList.begin(); glit != label->ghostlabelList.end(); glit++)
				{ 
					if ( *glit != NULL )
					{
						if ( _stricmp((*glit)->name.c_str(), optional) == NULL || _stricmp((*glit)->labelGroup.c_str(), optional) == NULL )
						{
							label->setText( text, false, 2, (*glit) );
						}
					}
					else
					{
						break;
					}
				}
			}
			else if ( textEditBox )
			{
				label->setText( text, false, 1);
			}
			else
			{
				label->setText(((numTokens >= 2) ? optional : ""), true);
			}
		}
	}
}

// !LabelToggle <label> [<overlaylabel>]
void BangToggle(HWND hwndCaller, LPCSTR pszArgs)
{
	bool textEditBox = false;
	char name[64] = {0};
	char optional[64] = {0};
	char *buffers[] = {name, optional};

	int numTokens = LCTokenize(pszArgs, buffers, 2, NULL);

	if ( _stricmp(optional, ".editbox") == NULL )
		textEditBox = true;

	Label *label = lookupLabel(name);
	
	if (label)
	{
		if ( numTokens >= 2 && !textEditBox )
		{
			for(GhostLabelListIterator glit = label->ghostlabelList.begin(); glit != label->ghostlabelList.end(); glit++)
			{ 
				if ( *glit != NULL )
				{
					if ( _stricmp((*glit)->name.c_str(), optional) == NULL || _stricmp((*glit)->labelGroup.c_str(), optional) == NULL )
					{
						if ( (*glit)->visible )
							label->hide(false, 2, (*glit));
						else
							label->show(false, 2, (*glit));
					}
				}
				else
				{
					break;
				}
			}
		}
		else if ( textEditBox )
		{
			if (label->isTextboxVisible())
				label->hide(false, 1);
			else
				label->show(false, 1);
		}
		else
		{
			if (label->isVisible())
				label->hide(true);
			else
				label->show(true);
		}
	}
	else
	{
		for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
		{
			if ( _stricmp((*lit)->labelGroup.c_str(), name) == NULL )
			{
				if ( numTokens >= 2 && !textEditBox )
				{
					for(GhostLabelListIterator glit = (*lit)->ghostlabelList.begin(); glit != (*lit)->ghostlabelList.end(); glit++)
					{ 
						if ( *glit != NULL )
						{
							if ( _stricmp((*glit)->name.c_str(), optional) == NULL || _stricmp((*glit)->labelGroup.c_str(), optional) == NULL )
							{
								if ( (*glit)->visible )
									(*lit)->hide(false, 2, (*glit));
								else
									(*lit)->show(false, 2, (*glit));
							}
						}
						else
						{
							break;
						}
					}
				}
				else if ( textEditBox )
				{
					if ((*lit)->isTextboxVisible())
						(*lit)->hide(false, 1);
					else
						(*lit)->show(false, 1);
				}
				else
				{
					if ((*lit)->isVisible())
						(*lit)->hide(true);
					else
						(*lit)->show(true);
				}
			}
		}
	}
}

// !LabelUpdate <label>
void BangUpdate(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = lookupLabel(pszArgs);

	if (label)
	{
		label->update();
	}
	else
	{
		for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
		{
			if ( _stricmp((*lit)->labelGroup.c_str(), pszArgs) == NULL )
				(*lit)->update();
		}
	}
}

// !LabelSetAlpha <label> <value> [<step>] [<delay>]
void BangSetAlpha(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], factor[16], step[16], delay[16];
	char *buffers[] = {name, factor, step, delay};

	int numTokens = LCTokenize(pszArgs, buffers, 4, NULL);

	if (numTokens >= 2)
	{
		Label *label = lookupLabel(name);
		
		if (label)
		{
			if ( label->getBox() )
				return;

			if ( numTokens < 4 )
			{
				label->setAlpha(atoi(factor));
			}
			else if ( numTokens == 4 )
			{
				label->setAlphaFade(atoi( factor ), atoi( step ), atoi( delay ));
			}
		}
		else
		{
			for ( LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++ )
			{
				if ( (*lit)->getBox() )
					continue;

				if ( _stricmp((*lit)->labelGroup.c_str(), name) == NULL )
				{
					if ( numTokens < 4 )
					{
						(*lit)->setAlpha(atoi(factor));
					}
					else if ( numTokens == 4 )
					{
						(*lit)->setAlphaFade(atoi( factor ), atoi( step ), atoi( delay ));
					}
				}
			}
		}
	}
}


// !LabelGhosted <label> <on|true, off|false, toggle>
void BangGhosted(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], option[32];
	char *buffers[] = {name, option};

	LCTokenize(pszArgs, buffers, 2, NULL);

	Label *label = lookupLabel(name);

	if (label)
	{
		if (_stricmp(option, "true") == NULL || _stricmp(option, "on") == NULL)
		{
			label->setGhosted(true);
		}
		else if (_stricmp(option, "false") == NULL || _stricmp(option, "off") == NULL)
		{
			label->setGhosted(false);
		}
		else if (_stricmp(option, "toggle") == NULL || _stricmp(option, "switch") == NULL)
		{
			label->setGhosted(!label->isGhosted());
		}
	}
	else
	{
		for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
		{
			if ( _stricmp((*lit)->labelGroup.c_str(), name) == NULL )
			{
				if (_stricmp(option, "true") == NULL || _stricmp(option, "on") == NULL)
				{
					(*lit)->setGhosted(true);
				}
				else if (_stricmp(option, "false") == NULL || _stricmp(option, "off") == NULL)
				{
					(*lit)->setGhosted(false);
				}
				else if (_stricmp(option, "toggle") == NULL || _stricmp(option, "switch") == NULL)
				{
					(*lit)->setGhosted(!(*lit)->isGhosted());
				}
			}
		}
	}
}

//Parse Current Evars In Bang
//!ParseEvars !anybang with evars escape code '%#' and TextEscape Sequences with %[...]%
void BangParseEvars(HWND hwndCaller, LPCSTR pszArgs)
{
	CString worker1 = pszArgs;
	CString worker2 = "";

	worker1.Replace("%#", "$");
	worker1.Replace("%|#", "$");

	while ( worker1.Find("%[") != -1 && worker1.Find("]%") != -1 )
	{
		worker2 = worker1;
		worker2 = worker2.Right(worker2.GetLength()-worker2.Find("%[")-1);
		worker2 = worker2.Left(worker2.Find("]%")+1);
		
		bool dynamic = true;
		if ( _stricmp(worker2, "[labelname]") == NULL || _stricmp(worker2, "[labeltext]") == NULL)
		{
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if( (*lit)->gethWnd() == hwndCaller)
				{
					worker1.Replace( ("%"+worker2+"%"), (systemInfo->processLabelText( (string)worker2, (*lit), &dynamic)).c_str() );
					break;
				}
			}
		}
		else
		{
			worker1.Replace( ("%"+worker2+"%"), (systemInfo->processLabelText( (string)worker2, (*labelList.begin()), &dynamic)).c_str() );
		}
	}

	xLSExecute(hwndCaller, worker1, SW_SHOWNORMAL);
}

//Set Evar
//!SetEvar "'$'evar'$'" "value"
void BangSetEvars(HWND hwndCaller, LPCSTR pszArgs)
{
	char evar[64], value[MAX_LINE_LENGTH];
	char *buffers[] = {evar};

	if (LCTokenize(pszArgs, buffers, 1, value) >= 1)
		PrefixedLSSetVariable( (string)evar, "", (string)value );
}

//Play Wav
//!PlaySound "Path to wav|mp3|mid"
void BangPlaySound(HWND hwndCaller, LPCSTR pszArgs)
{
	CString m_Path = pszArgs;
	m_Path.Remove('\"');
	m_Path.MakeLower();

	if (m_Path.Find(".wav") || m_Path.Find(".mp3") || m_Path.Find(".mid"))
	{
		HANDLE hFindHandle;
		WIN32_FIND_DATA FindData;

		hFindHandle = FindFirstFile(m_Path, &FindData);

		if (INVALID_HANDLE_VALUE != hFindHandle)
		{
			if ( !m_hWndSoundFile )
			{
				m_hWndSoundFile = MCIWndCreate(NULL, NULL, WS_MINIMIZE | MCIWNDF_NOMENU | MCIWNDF_NOERRORDLG, m_Path);
			}
			else
			{
				MCIWndStop(m_hWndSoundFile);

				MCIWndOpen(m_hWndSoundFile, (LPCTSTR) m_Path, NULL );
			}
			
			MCIWndPlay(m_hWndSoundFile);

			SetTimer(messageHandler, TIMER_PLAYSOUND, 1000, 0);
		}

		FindClose(hFindHandle);
	}
	else if (m_hWndSoundFile != NULL)
	{
		MCIWndStop(m_hWndSoundFile);
	}
}

void BangTexteditBoxExecute(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], action[MAX_LINE_LENGTH];
	char *buffers[] = {name};

	LCTokenize(pszArgs, buffers, 1, action);

	char buf[MAX_LINE_LENGTH] = {0};
	CString worker = "";

	Label *label = lookupLabel(name);
	if(label)
	{
		if ( label->getTexteditBoxhWnd() )
		{
			GetWindowText(label->getTexteditBoxhWnd(), buf, sizeof(buf));

			worker = action;
			worker.Replace("%[textinput]%", buf);

			xLSExecute(hwndCaller, worker, SW_SHOWNORMAL);
		}
	}
	else
	{
		LabelListIterator lit;
		for( lit = labelList.begin(); lit != labelList.end(); lit++ )
		{
			if( _stricmp((*lit)->labelGroup.c_str(), name) == NULL )
			{
				if ( (*lit)->getTexteditBoxhWnd() )
				{
					GetWindowText((*lit)->getTexteditBoxhWnd(), buf, sizeof(buf));

					worker = action;
					worker.Replace("%[textinput]%", buf);

					xLSExecute(hwndCaller, worker, SW_SHOWNORMAL);
				}
			}
		}
	}

}
